#
# Copyright (c) 2019-2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unicodedata
from dataclasses import dataclass
from typing import Iterable, List, NamedTuple, Union

import torch


def identity(x, **kw):
    return x, [], x


class SequentialField(NamedTuple):
    value: Union[torch.tensor, List[int]]
    length: Union[torch.tensor, int]
    limited: Union[torch.tensor, List[int]]
    feature: Union[torch.tensor, List[List[int]], None]


# Feature is defined per token
# Each field contains a list of possible values for that feature
@dataclass
class Feature:
    type_id: List[int] = None
    type_prob: List[float] = None

    def __mul__(self, n):
        return [self for _ in range(n)]

    def flatten(self):
        result = []
        for field in VALID_FEATURE_FIELDS:
            field_val = getattr(self, field)
            if field_val:
                result += field_val
        return result


VALID_FEATURE_FIELDS = tuple(Feature.__annotations__.keys())


def get_pad_feature(feature_fields, ned_features_default_val, ned_features_size):
    # return None if not using NED
    pad_feature = None
    if len(feature_fields):
        pad_feature = Feature()
        for i, field in enumerate(feature_fields):
            assert field in VALID_FEATURE_FIELDS
            setattr(pad_feature, field, [ned_features_default_val[i]] * ned_features_size[i])
    return pad_feature


class Example(object):
    """
    Contains all fields of a train/dev/test example in text form, alongside their NED features
    both in text form (`*_plus_types` fields) and in embedding_id form (`*_feature`)
    """

    def __init__(
        self,
        example_id: str,
        context: str,
        context_feature: List[Feature],
        context_plus_types: str,
        question: str,
        question_feature: List[Feature],
        question_plus_types: str,
        answer: str,
    ):

        self.example_id = example_id
        self.context = context
        self.context_feature = context_feature
        self.context_plus_types = context_plus_types
        self.question = question
        self.question_feature = question_feature
        self.question_plus_types = question_plus_types
        self.answer = answer

    @staticmethod
    def from_raw(example_id: str, context: str, question: str, answer: str, preprocess=identity, lower=False):
        args = [example_id]
        answer = unicodedata.normalize('NFD', answer)

        for argname, arg in (('context', context), ('question', question), ('answer', answer)):
            arg = unicodedata.normalize('NFD', arg)
            if lower:
                arg = arg.lower()

            sentence, features, sentence_plus_types = preprocess(arg.rstrip('\n'), field_name=argname, answer=answer)

            args.append(sentence)

            if argname != 'answer':
                args.append(features)

            if argname == 'context':
                context_plus_types = sentence_plus_types
                args.append(context_plus_types)
            elif argname == 'question':
                question_plus_types = sentence_plus_types
                args.append(question_plus_types)

        return Example(*args)


class NumericalizedExamples(NamedTuple):
    """
    Conatains a batch of numericalized (i.e. tokenized and converted to token ids) examples, potentially of size 1
    """

    example_id: List[str]
    context: SequentialField
    answer: SequentialField

    @staticmethod
    def from_examples(examples: Iterable[Example], numericalizer):
        assert all(isinstance(ex.example_id, str) for ex in examples)
        numericalized_examples = []
        args = numericalizer.args

        if args.no_separator:
            sep_token = ' '
            pad_feature = []
        else:
            sep_token = ' ' + numericalizer.sep_token + ' '
            pad_feature = [get_pad_feature(args.ned_features, args.ned_features_default_val, args.ned_features_size)]

        # we keep the result of concatenation of question and context fields in these arrays temporarily. The numericalized versions will live on in self.context
        all_context_plus_questions = []
        all_context_plus_question_with_types = []
        all_context_plus_question_features = []

        for ex in examples:
            # create context_plus_question fields by concatenating context and question fields
            # if question is empty, don't append anything
            context_plus_question = ex.context + sep_token + ex.question if len(ex.question) else ex.context
            all_context_plus_questions.append(context_plus_question)
            context_plus_question_with_types = (
                ex.context_plus_types + sep_token + ex.question_plus_types
                if len(ex.question_plus_types)
                else ex.context_plus_types
            )
            all_context_plus_question_with_types.append(context_plus_question_with_types)
            # concatenate question and context features with a separator, but no need for a separator if there are no features to begin with
            context_plus_question_feature = (
                ex.context_feature + pad_feature + ex.question_feature
                if len(ex.question_feature) + len(ex.context_feature) > 0
                else []
            )
            all_context_plus_question_features.append(context_plus_question_feature)

        if args.add_types_to_text == 'no':
            tokenized_contexts = numericalizer.encode_batch(
                all_context_plus_questions, field_name='context', features=[a for a in all_context_plus_question_features if a]
            )
        else:
            tokenized_contexts = numericalizer.encode_batch(all_context_plus_question_with_types, field_name='context')

        # TODO remove double attempts at context tokenization
        if getattr(examples, 'is_classification', False):
            assert args.add_types_to_text != 'insert'
            tokenized_answers = numericalizer.process_classification_labels(
                all_context_plus_questions, all_context_plus_question_with_types, [ex.answer for ex in examples]
            )
        else:
            tokenized_answers = numericalizer.encode_batch([ex.answer for ex in examples], field_name='answer')

        for i in range(len(examples)):
            numericalized_examples.append(
                NumericalizedExamples([examples[i].example_id], tokenized_contexts[i], tokenized_answers[i])
            )
        return numericalized_examples

    @staticmethod
    def collate_batches(batches: Iterable['NumericalizedExamples'], numericalizer, device):
        example_id = []

        context_values, context_lengths, context_limiteds, context_features = [], [], [], []
        answer_values, answer_lengths, answer_limiteds = [], [], []

        for batch in batches:
            example_id.append(batch.example_id[0])
            context_values.append(torch.tensor(batch.context.value, device=device))
            context_lengths.append(torch.tensor(batch.context.length, device=device))
            context_limiteds.append(torch.tensor(batch.context.limited, device=device))
            if batch.context.feature:
                context_features.append(torch.tensor(batch.context.feature, device=device))

            answer_values.append(torch.tensor(batch.answer.value, device=device))
            answer_lengths.append(torch.tensor(batch.answer.length, device=device))
            answer_limiteds.append(torch.tensor(batch.answer.limited, device=device))

        context_values = numericalizer.pad(context_values, pad_id=numericalizer.pad_id)
        context_limiteds = numericalizer.pad(context_limiteds, pad_id=numericalizer.decoder_pad_id)
        context_lengths = torch.stack(context_lengths, dim=0)

        if context_features:
            context_features = numericalizer.pad(context_features, pad_id=numericalizer.args.db_unk_id)

        answer_values = numericalizer.pad(answer_values, pad_id=numericalizer.pad_id)
        answer_limiteds = numericalizer.pad(answer_limiteds, pad_id=numericalizer.decoder_pad_id)
        answer_lengths = torch.stack(answer_lengths, dim=0)

        context = SequentialField(
            value=context_values, length=context_lengths, limited=context_limiteds, feature=context_features
        )

        answer = SequentialField(value=answer_values, length=answer_lengths, limited=answer_limiteds, feature=None)

        return NumericalizedExamples(example_id=example_id, context=context, answer=answer)



#
# Copyright (c) 2019-2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import json
import logging
import os
import re
from collections import Counter, defaultdict
from typing import List, Tuple

from pathos import multiprocessing
from torch.nn.utils.rnn import pad_sequence
from transformers import (
    SPIECE_UNDERLINE,
    AutoTokenizer,
    BertTokenizer,
    BertTokenizerFast,
    GPT2Tokenizer,
    GPT2TokenizerFast,
    M2M100Tokenizer,
    MarianConfig,
    MarianTokenizer,
    MBart50Tokenizer,
    MBart50TokenizerFast,
    T5Config,
    T5Tokenizer,
    T5TokenizerFast,
    XLMRobertaTokenizer,
    XLMRobertaTokenizerFast,
)

from ..util import get_devices
from .decoder_vocab import DecoderVocabulary
from .example import SequentialField, get_pad_feature

logger = logging.getLogger(__name__)

# not all tokenizers respect whitespace in the input or honor do_basic_tokenize=False
# for those, we need to use the slow tokenizers or we'll get messed up thingtalk output
ALLOWED_FAST_TOKENIZERS = {
    'facebook/bart-base',
    'facebook/bart-large',
    'sshleifer/bart-tiny-random',
}

# known NOT to work:
# - all the BERT models
# - all the XLM-R models
#
# mBART, t5, and mt5 models work when preprocessing, because they're SPM/RoBERTa-based so they respect
# whitespace, but the fast tokenizers treat special tokens differently than the slow ones
# and drop whitespace before special tokens, which breaks
ALLOWED_FAST_TOKENIZERS_IF_PREPROCESSING = {
    'facebook/mbart-large-cc25',
    'sshleifer/tiny-mbart',
    'google/mt5-small',
    'google/mt5-base',
    'google/mt5-large',
    'google/mt5-xl',
    'google/mt5-xxl',
}

# for input batches smaller than this value, multiprocessing will not be used due to its overhead
MULTIPROCESSING_THRESHOLD = 5000


class TransformerNumericalizer(object):
    """
    Numericalizer that uses Tokenizers from huggingface's transformers library.
    """

    _special_tokens_to_word_map: List[Tuple[str, str]]
    _special_tokens_to_word_regexes: List[Tuple[re.Pattern, str]]
    _special_tokens_to_token_regexes: List[Tuple[re.Pattern, str]]

    def __init__(
        self, pretrained_tokenizer, args, max_generative_vocab, config, src_lang, tgt_lang, vocab_sets, tasks, save_dir=None
    ):
        """
        If `save_dir` is None, initializes a new Numericalizer and optionally adds new words to its vocabulary, otherwise,
        loads from `save_dir`
        """
        self._pretrained_name = pretrained_tokenizer
        self.max_generative_vocab = max_generative_vocab
        self._cache = args.embeddings
        self._tokenizer = None

        self._preprocess_special_tokens = args.preprocess_special_tokens

        # map a token to a space-separated sequence of words
        self._special_tokens_to_word_map = []
        # same, but the token is a regular expression matching that token using \b
        self._special_tokens_to_word_regexes = []
        # map a space-separated sequence of words to a token
        self._special_tokens_to_token_regexes = []

        self.args = args

        self._init_tokenizer(save_dir, config, src_lang, tgt_lang)

        if save_dir is not None:
            logger.info(f'Loading the accompanying numericalizer from {save_dir}')
            self.load_extras(save_dir)
        else:
            logger.info('Building vocabulary')
            self.build_vocab(vocab_sets, tasks)

        self._init_token_ids()
        self._init_decoder_vocab()

    @property
    def vocab(self):
        return self._tokenizer

    @property
    def num_tokens(self):
        return len(self._tokenizer)

    @property
    def decoder_pad_id(self):
        if self.max_generative_vocab is not None:
            return self.decoder_vocab.pad_idx
        else:
            return self.pad_id

    def _use_fast(self):
        if self.args.no_fast_tokenizer:
            return False
        if self.args.force_fast_tokenizer:
            return True
        return self._pretrained_name in ALLOWED_FAST_TOKENIZERS or (
            self._preprocess_special_tokens and self._pretrained_name in ALLOWED_FAST_TOKENIZERS_IF_PREPROCESSING
        )

    def _init_tokenizer(self, save_dir, config, src_lang, tgt_lang):
        """
        Initializes the `self._tokenizer` object, but not the rest.
        """
        tokenizer_args = {
            'do_lower_case': False,
            'do_basic_tokenize': False,
            'cache_dir': self._cache,
            'use_fast': self._use_fast(),
            'src_lang': src_lang,
            'tgt_lang': tgt_lang,
        }
        if save_dir is not None:
            tokenizer_args.update({'pretrained_model_name_or_path': save_dir, 'config': config})
        else:
            tokenizer_args.update({'pretrained_model_name_or_path': self._pretrained_name})

        self._tokenizer = AutoTokenizer.from_pretrained(**tokenizer_args)

        # some tokenizers like Mbart do not set src_lang and tgt_lan when initialized; take care of it here
        self._tokenizer.src_lang = src_lang
        self._tokenizer.tgt_lang = tgt_lang

        # define input prefix to add before every input text
        input_prefix = ''
        if isinstance(config, MarianConfig) and tgt_lang:
            input_prefix = f'>>{tgt_lang}<< '
        elif isinstance(config, T5Config):
            t5_task = f'translation_{src_lang}_to_{tgt_lang}'
            # TODO add support for summarization
            # t5_task = 'summarization'
            input_prefix = config.task_specific_params[t5_task]['prefix']

        self.input_prefix = input_prefix

        # We only include the base tokenizers since `isinstance` checks for inheritance
        if isinstance(self._tokenizer, (BertTokenizer, BertTokenizerFast)):
            self._tokenizer.is_piece_fn = lambda wp: wp.startswith('##')
        elif isinstance(
            self._tokenizer,
            (
                XLMRobertaTokenizer,
                XLMRobertaTokenizerFast,
                T5Tokenizer,
                T5TokenizerFast,
                MBart50Tokenizer,
                MBart50TokenizerFast,
                MarianTokenizer,
                M2M100Tokenizer,
            ),
        ):
            self._tokenizer.is_piece_fn = lambda wp: not wp.startswith(SPIECE_UNDERLINE)
        elif isinstance(self._tokenizer, (GPT2Tokenizer, GPT2TokenizerFast)):
            self._tokenizer.is_piece_fn = lambda wp: not wp.startswith('Ġ')

        # make sure we assigned is_piece_fn
        assert self._tokenizer.is_piece_fn

    def load_extras(self, save_dir):
        if self.max_generative_vocab is not None:
            with open(os.path.join(save_dir, 'decoder-vocab.txt'), 'r') as fp:
                self._decoder_words = [
                    (line.rstrip('\n'), self._tokenizer.convert_tokens_to_ids(line.rstrip('\n'))) for line in fp
                ]
        try:
            with open(os.path.join(save_dir, 'special-token-preprocessing.json')) as fp:
                self._special_tokens_to_word_map = json.load(fp)
            self._build_special_tokens_regexes()
        except FileNotFoundError:
            pass

    def pad(self, batch, pad_id):
        """
        batch: a List of List of integers
        """
        # TODO account for left padding models
        return pad_sequence(batch, padding_value=pad_id, batch_first=True)

    def save(self, save_dir):
        self._tokenizer.save_pretrained(save_dir)
        if self.max_generative_vocab is not None:
            with open(os.path.join(save_dir, 'decoder-vocab.txt'), 'w') as fp:
                for word, _full_idx in self._decoder_words:
                    fp.write(word + '\n')
        if len(self._special_tokens_to_word_map) > 0:
            with open(os.path.join(save_dir, 'special-token-preprocessing.json'), 'w') as fp:
                json.dump(self._special_tokens_to_word_map, fp)

    def build_vocab(self, vocab_sets, tasks):
        special_tokens = []
        for task in tasks:
            special_tokens += list(task.special_tokens)
        special_tokens.sort()

        if self._preprocess_special_tokens:
            self._build_special_tokens_maps(special_tokens)
            self._build_special_tokens_regexes()
        else:
            # add the special tokens directly to the tokenizer
            self._tokenizer.add_tokens(special_tokens)

        # add entity boundary special tokens
        if self.args.add_types_to_text != 'no':
            self._tokenizer.add_tokens(['<e>', '</e>'])

        # add special tokens for ambig_qa task
        if self.args.train_task_names == 'ambig_qa':
            self._tokenizer.add_tokens(['<q>', '<p>', '<u>'])

        existing_special_tokens = self._tokenizer.special_tokens_map
        # add separator if it doesn't exist. It will be used to concatenate context and question
        if 'sep_token' not in existing_special_tokens:
            self._tokenizer.add_special_tokens({'sep_token': existing_special_tokens.get('sep_token', '</s>')})

        if self.max_generative_vocab is not None:
            # do a pass over all the data in the dataset
            # in this pass, we
            # 1) tokenize everything, to ensure we account for all added tokens
            # 2) we construct a counter of wordpieces in the answers, for the decoder vocabulary
            decoder_words = Counter()
            for dataset in vocab_sets:
                for example in dataset:
                    decoder_words.update(self._tokenizer.tokenize(example.context))
                    decoder_words.update(self._tokenizer.tokenize(example.question))
                    decoder_words.update(self._tokenizer.tokenize(example.answer))

            # add the required special tokens, if not present already
            # note: if the tokens are not present, it means they are not used natively
            # by the model, so we can pick our favorite token
            if 'bos_token' not in existing_special_tokens:
                self._tokenizer.add_special_tokens({'bos_token': existing_special_tokens.get('cls_token', '<s>')})
            if 'eos_token' not in existing_special_tokens:
                self._tokenizer.add_special_tokens({'eos_token': existing_special_tokens.get('sep_token', '</s>')})
            if 'pad_token' not in existing_special_tokens:
                self._tokenizer.add_special_tokens({'pad_token': '<pad>'})
            if 'unk_token' not in existing_special_tokens:
                self._tokenizer.add_special_tokens({'unk_token': '<unk>'})
            self._decoder_words = [
                (self._tokenizer.bos_token, self._tokenizer.bos_token_id),
                (self._tokenizer.eos_token, self._tokenizer.eos_token_id),
                (self._tokenizer.pad_token, self._tokenizer.pad_token_id),
                (self._tokenizer.unk_token, self._tokenizer.unk_token_id),
            ] + [
                (word, self._tokenizer.convert_tokens_to_ids(word))
                for word, _freq in decoder_words.most_common(self.max_generative_vocab)
            ]

    def grow_vocab(self, tasks):
        if self._preprocess_special_tokens:
            # if we're preprocessing special tokens, we cannot extend the vocabulary
            # if the vocabulary was incomplete during training, tough luck, those words will be subwords all the way
            # (what do you expect?)
            return

        # add the new special tokens from the task
        for task in tasks:
            self._tokenizer.add_tokens(list(task.special_tokens))

    def _build_special_tokens_maps(self, special_tokens):
        # we automatically construct the mapping from special tokens to the shortest unambiguous
        # sequence of word-like things

        processed_tokens = dict()
        # first, split each token into words
        for original_token in special_tokens:
            token = original_token
            prefix = None
            if token.startswith('@'):
                prefix = '@ '
                token = token[1:]
            elif token.startswith('^^'):
                prefix = '^^ '
                token = token[2:]

            # split the token into words
            parts = re.split('[:._-]', token)
            if not prefix:
                # if we don't have a prefix, use the first word as prefix
                assert len(parts) >= 2
                prefix = parts[0] + ' '
                parts = parts[1:]
            processed_tokens[original_token] = (prefix, parts)

        # words -> token(s) (multiple tokens if a sequence of words is ambiguous)
        assignment = defaultdict(list)
        # token -> words
        reverse_mapping = dict()

        # now greedily assign each token to the shortest end that is not ambiguous
        for original_token, (prefix, parts) in processed_tokens.items():
            for i in range(len(parts) - 1, -1, -1):
                attempt = prefix + ' '.join(parts[i:])
                if attempt in assignment:
                    # ambiguous, extend everything in the current assignment by one word
                    for ambiguous_token in assignment[attempt]:
                        # list of sequences of words mapping to this token
                        # the first one is the one we have chosen so far, and the others are ambiguous
                        word_sequences = reverse_mapping[ambiguous_token]
                        if word_sequences[0] != attempt:
                            # ambiguous_token is already choosing a word sequence that is not ambiguous with
                            # original_token, nothing to do, other than to know original_token will need to
                            # be extended
                            continue
                        # extend ambiguous_token by one
                        ambiguous_prefix, ambiguous_parts = processed_tokens[ambiguous_token]
                        # assert we still have one word to use to disambiguate
                        # this works as long as the tokens are not suffix of one another
                        assert len(word_sequences) < len(ambiguous_parts), (original_token, ambiguous_token, word_sequences)
                        new_words = ambiguous_prefix + ' '.join(
                            ambiguous_parts[len(ambiguous_parts) - len(word_sequences) - 1 :]
                        )
                        word_sequences.insert(0, new_words)
                        # before original_token, ambiguous_token was not ambiguous with any token already
                        # assigned, so it cannot be ambiguous after we made it longer
                        assert new_words not in assignment
                        assignment[new_words] = [ambiguous_token]

                    # mark that attempt is an ambiguous suffix of original_token
                    assignment[attempt].append(original_token)

                    # don't assign original_token at this step, wait until the next loop cycle
                    # at the next loop, we'll try again with a longer suffix of original_token
                    # that way, we check if the token is still ambiguous after we extended everything
                    # else by one word
                else:
                    # yay not ambiguous, time to assign it

                    # construct all word sequences, from the one we chose to the end
                    word_sequences = [prefix + ' '.join(parts[j:]) for j in range(i, len(parts))]
                    assignment[attempt] = [original_token]
                    reverse_mapping[original_token] = word_sequences
                    break

        # okay we have assigned everything, time to clean up
        for token, word_sequences in reverse_mapping.items():
            self._special_tokens_to_word_map.append((token, word_sequences[0]))

    def _build_special_tokens_regexes(self):
        for token, words in self._special_tokens_to_word_map:
            # match requiring (at the beginning of the string or preceded by a space (positive lookbehind))
            # and (at the end of the string or followed by a space (positive lookahead))
            word_re = re.compile("(^|(?<= ))" + re.escape(words) + "($|(?= ))")
            self._special_tokens_to_token_regexes.append((word_re, token))
            token_re = re.compile("(^|(?<= ))" + re.escape(token) + "(^|(?= ))")
            self._special_tokens_to_word_regexes.append((token_re, words))

    def _init_token_ids(self):
        self.pad_first = self._tokenizer.padding_side == 'left'

        self.init_token = self._tokenizer.bos_token
        self.eos_token = self._tokenizer.eos_token
        self.unk_token = self._tokenizer.unk_token
        self.pad_token = self._tokenizer.pad_token
        self.mask_token = self._tokenizer.mask_token
        self.cls_token = self._tokenizer.cls_token
        self.sep_token = self._tokenizer.sep_token

        self.init_id = self._tokenizer.bos_token_id
        self.eos_id = self._tokenizer.eos_token_id
        self.unk_id = self._tokenizer.unk_token_id
        self.pad_id = self._tokenizer.pad_token_id
        self.mask_id = self._tokenizer.mask_token_id
        self.cls_id = self._tokenizer.cls_token_id
        self.sep_id = self._tokenizer.sep_token_id

        # pad_id for answer tokens (different than context/ question pad_id for token classification tasks)
        self.answer_pad_id = self.pad_id

    def _init_decoder_vocab(self):
        if self.max_generative_vocab is not None:
            self.generative_vocab_size = len(self._decoder_words)

            self.decoder_vocab = DecoderVocabulary(
                self._decoder_words, self._tokenizer, pad_token=self.pad_token, eos_token=self.eos_token
            )
        else:
            self.generative_vocab_size = len(self._tokenizer)
            self.decoder_vocab = None

    def get_num_special_tokens(self, special_tokens_mask):
        num_prefix_special_tokens, num_suffix_special_tokens = 0, 0
        i = 0
        while i < len(special_tokens_mask):
            if special_tokens_mask[i] == 1:
                num_prefix_special_tokens += 1
                i += 1
            else:
                break
        i = len(special_tokens_mask) - 1
        while i >= 0:
            if special_tokens_mask[i] == 1:
                num_suffix_special_tokens += 1
                i -= 1
            else:
                break

        return num_prefix_special_tokens, num_suffix_special_tokens

    def process_classification_labels(self, all_context_plus_questions, all_context_plus_question_with_types, all_answers):
        def tokenize_and_align_labels(all_sequences, all_sequences_with_types, all_labels, label_all_tokens=True):
            tokenized_inputs = self._tokenizer.batch_encode_plus(
                all_sequences,
                padding=False,
                truncation=True,
                # We use this argument because the texts in our dataset are lists of words (with a label for each word).
                is_split_into_words=True,
            )
            tokenized_inputs_with_types = self._tokenizer.batch_encode_plus(
                all_sequences_with_types,
                padding=False,
                truncation=True,
                # We use this argument because the texts in our dataset are lists of words (with a label for each word).
                is_split_into_words=True,
            )

            all_processed_labels = []
            for i, label in enumerate(all_labels):
                word_ids = tokenized_inputs.word_ids(batch_index=i)
                word_ids_with_types = tokenized_inputs_with_types.word_ids(batch_index=i)

                # find types size
                size_diff = len(word_ids_with_types) - len(word_ids)

                previous_word_idx = None
                label_ids = []
                for word_id in word_ids:
                    # special tokens's word_id is None
                    if word_id is None:
                        label_ids.append(self.answer_pad_id)
                    # set the label for the first subword of each word
                    elif word_id != previous_word_idx:
                        label_ids.append(label[word_id])
                    # process next subwords
                    else:
                        label_ids.append(label[word_id] if label_all_tokens else self.answer_pad_id)
                    previous_word_idx = word_id

                # extend labels for types
                label_ids.extend([self.answer_pad_id] * size_diff)

                all_processed_labels.append(label_ids)

            return all_processed_labels

        # align labels
        assert all(
            [len(cpq.split(" ")) == len(answer.split(" ")) for cpq, answer in zip(all_context_plus_questions, all_answers)]
        )

        tokenized_answers = tokenize_and_align_labels(
            [cpq.split(" ") for cpq in all_context_plus_questions],
            [cpqt.split(" ") for cpqt in all_context_plus_question_with_types],
            [list(map(lambda token: int(token), ans.split(" "))) for ans in all_answers],
        )

        batch_decoder_numerical = []
        if self.decoder_vocab:
            for i in range(len(tokenized_answers)):
                batch_decoder_numerical.append(self.decoder_vocab.encode(tokenized_answers[i]))
        else:
            batch_decoder_numerical = [[]] * len(tokenized_answers)

        answer_sequential_fields = []
        for i in range(len(tokenized_answers)):
            answer_sequential_fields.append(
                SequentialField(
                    value=tokenized_answers[i],
                    length=len(tokenized_answers[i]),
                    limited=batch_decoder_numerical[i],
                    feature=None,
                )
            )

        return answer_sequential_fields

    def encode_batch(self, sentences: List[str], field_name, features=None) -> List[SequentialField]:
        """
        Batched version of `encode_single()`. Uses multiprocessing on all CPU cores for preprocessing,
        and multithreading for tokenization if a `FastTokenizer` is used
        Inputs:
            sentences: a list of sentences to encode
            field_name: text field name (options: context, question, answer)
            features: for each sentence we have a list of features per token (used for NED)
        """
        # We need to set this so that `tokenizers` package does not complain about detecting forks.
        os.environ['TOKENIZERS_PARALLELISM'] = "true"

        if features is None:
            features = []
            extract_word_pieces = False
        else:
            assert all([len(sentence.split()) == len(feature) for sentence, feature in zip(sentences, features)])
            extract_word_pieces = True

        batch_size = len(sentences)

        if field_name != 'answer':
            sentences = [self.input_prefix + sent for sent in sentences]

        if self._preprocess_special_tokens:
            if len(sentences) > MULTIPROCESSING_THRESHOLD:
                multiprocessing_factor = multiprocessing.cpu_count() // len(get_devices(self.args.devices))
                logger.info('multiprocessing factor for special token preprocessing is %d', multiprocessing_factor)
                with multiprocessing.Pool(multiprocessing_factor) as p:
                    sentences, index2expansions = list(
                        zip(
                            *p.map(
                                functools.partial(self._apply_special_token_preprocessing, return_idx2exp=bool(len(features))),
                                sentences,
                            )
                        )
                    )
            else:
                sentences, index2expansions = list(
                    zip(
                        *map(
                            functools.partial(self._apply_special_token_preprocessing, return_idx2exp=bool(len(features))),
                            sentences,
                        )
                    )
                )

            all_input_features = []
            if features:
                for i, (sentence, index2expansion) in enumerate(zip(sentences, index2expansions)):
                    feat = features[i]
                    new_feat = []
                    keys = set(index2expansion.keys())
                    for j in range(len(feat)):
                        repeat = 1
                        if j in keys:
                            repeat = index2expansion[j]

                        new_feat.extend(feat[j] * repeat)

                    assert len(new_feat) == len(sentence.split(' '))

                    all_input_features.append(new_feat)

            features = all_input_features

        # batch_encode_plus for fast tokenizers returns tokenized text
        # whereas slow version do not. We breakdown slow tokenization into two steps
        # extract tokenized text first, use that to adjust features
        # then pass tokenized text to `_batch_prepare_for_model`
        def do_fast_tokenization(extract_word_pieces):
            all_wp_tokenized = []
            batch_encoded = self._tokenizer.batch_encode_plus(
                list(sentences),
                add_special_tokens=True,
                max_length=None,
                return_length=True,
                return_attention_mask=False,
                return_special_tokens_mask=True,
            )
            if extract_word_pieces:
                for encoding in batch_encoded.encodings:
                    # remove special tokens
                    num_prefix_special_tokens, num_suffix_special_tokens = self.get_num_special_tokens(
                        encoding.special_tokens_mask
                    )
                    wp_tokens = encoding.tokens[num_prefix_special_tokens:-num_suffix_special_tokens]
                    all_wp_tokenized.append(wp_tokens)

            return batch_encoded, all_wp_tokenized

        def do_slow_tokenization(extract_word_pieces):
            all_input_ids = []
            all_wp_tokenized = []
            if extract_word_pieces:
                for i in range(batch_size):
                    text = sentences[i]
                    wp_tokenized = self._tokenizer.tokenize(text)
                    all_wp_tokenized.append(wp_tokenized)

                    # None indicates encoding single instance not paired inputs
                    all_input_ids.append((self._tokenizer.convert_tokens_to_ids(wp_tokenized), None))

                batch_encoded = self._tokenizer._batch_prepare_for_model(
                    all_input_ids,
                    add_special_tokens=True,
                    max_length=None,
                    return_length=True,
                    return_attention_mask=False,
                    return_special_tokens_mask=True,
                )
            else:
                batch_encoded = self._tokenizer.batch_encode_plus(
                    list(sentences),
                    add_special_tokens=True,
                    max_length=None,
                    return_length=True,
                    return_attention_mask=False,
                    return_special_tokens_mask=True,
                )
            return batch_encoded, all_wp_tokenized

        if self._use_fast():
            if field_name == 'answer':
                with self._tokenizer.as_target_tokenizer():
                    batch_encoded, all_wp_tokenized = do_fast_tokenization(extract_word_pieces)
            else:
                batch_encoded, all_wp_tokenized = do_fast_tokenization(extract_word_pieces)

        else:
            if field_name == 'answer':
                with self._tokenizer.as_target_tokenizer():
                    batch_encoded, all_wp_tokenized = do_slow_tokenization(extract_word_pieces)
            else:
                batch_encoded, all_wp_tokenized = do_slow_tokenization(extract_word_pieces)

        all_input_features = []

        if features:
            for i in range(batch_size):
                wp_features = []
                wp_tokenized = all_wp_tokenized[i]

                feat = features[i]

                # first token is always not a piece
                is_wp = [0] + [int(self._tokenizer.is_piece_fn(wp)) for wp in wp_tokenized[1:]]
                k = -1
                for j, wp in enumerate(wp_tokenized):
                    if not is_wp[j]:
                        k += 1
                    wp_features.append(feat[k])

                assert len(wp_tokenized) == len(wp_features)

                all_input_features.append(wp_features)

        features = all_input_features

        batch_special_tokens_mask = batch_encoded.special_tokens_mask

        batch_features = []

        if features:
            for i in range(batch_size):
                feat = features[i]
                special_tokens_mask = batch_special_tokens_mask[i]
                num_prefix_special_tokens, num_suffix_special_tokens = self.get_num_special_tokens(special_tokens_mask)

                pad_feat = get_pad_feature(
                    self.args.ned_features, self.args.ned_features_default_val, self.args.ned_features_size
                )
                feat = [pad_feat] * num_prefix_special_tokens + feat + [pad_feat] * num_suffix_special_tokens

                batch_features.append(feat)

        batch_numerical = batch_encoded.input_ids
        batch_length = batch_encoded.length

        batch_decoder_numerical = []
        if self.decoder_vocab:
            for i in range(batch_size):
                batch_decoder_numerical.append(self.decoder_vocab.encode(batch_numerical[i]))
        else:
            batch_decoder_numerical = [[]] * len(batch_numerical)

        sequential_fields = []
        for i in range(batch_size):
            if features:
                feature = [feat.flatten() for feat in batch_features[i]]
                assert len(batch_numerical[i]) == len(feature)
            else:
                feature = None

            sequential_fields.append(
                SequentialField(
                    value=batch_numerical[i], length=batch_length[i], limited=batch_decoder_numerical[i], feature=feature
                )
            )
        return sequential_fields

    def _apply_special_token_preprocessing(self, sentence, return_idx2exp=False):
        index2expansion = {}
        if return_idx2exp:
            for i, (regex, replacement) in enumerate(self._special_tokens_to_word_regexes):
                for match in regex.finditer(sentence):
                    idx = match.span()[0]
                    tokens_before_idx = len(sentence[:idx].split(' '))
                    index2expansion[tokens_before_idx] = len(replacement.split(' '))
        for i, (regex, replacement) in enumerate(self._special_tokens_to_word_regexes):
            sentence = regex.sub(replacement, sentence)
        return sentence, index2expansion

    def _undo_special_token_preprocessing(self, sentence):
        for regex, replacement in self._special_tokens_to_token_regexes:
            sentence = regex.sub(replacement, sentence)
        return sentence

    def reverse(self, batch, field_name, skip_special_tokens=True):
        output = []
        use_source_tokenizer = True
        if field_name == 'answer':
            use_source_tokenizer = False
        for x in self._tokenizer.batch_decode(
            batch,
            skip_special_tokens=skip_special_tokens,
            clean_up_tokenization_spaces=False,
            use_source_tokenizer=use_source_tokenizer,
        ):
            if self._preprocess_special_tokens:
                x = self._undo_special_token_preprocessing(x)
            output.append(x)
        return output

    def convert_ids_to_tokens(self, batch, skip_special_tokens):
        output = []
        for ids in batch:
            x = self._tokenizer.convert_ids_to_tokens(ids, skip_special_tokens=skip_special_tokens)
            if self._preprocess_special_tokens:
                x = self._undo_special_token_preprocessing(x)
            output.append(x)
        return output



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import re
import unicodedata

import nltk
from nltk.corpus import stopwords

nltk.download('stopwords', quiet=True)

BANNED_PHRASES = set(
    stopwords.words('english')
    + [
        'music',
        'musics',
        'name',
        'names',
        'want',
        'wants',
        'album',
        'albums',
        'please',
        'who',
        'show me',
        'tell me',
        'find me',
        'sing',
        'sang',
        'play',
        'play me',
        'plays',
        'track',
        'tracks',
        'song',
        'songs',
        'record',
        'records',
        'recordings',
        'album',
        'url',
        'mount to',
        'something',
        'get',
        'selections',
        'pages',
        'isbn',
        'isbn numbers',
        'average rating',
        'count',
        'yesterday',
        'before today',
        'i need to know',
        'resume',
        'resumes',
        'the',
        'search for me',
        'search',
        'searches',
        'yes',
        'yeah',
        'popular',
        'trouble',
        'go',
        'millisecond',
        'good music',
        'hear music',
        'h',
        'm',
        's',
        'd',
        'y',
        'am',
        'pm',
        'min',
        'sec',
        'hour',
        'year',
        'month',
        'day',
        'us',
        'we',
        'who',
        'what',
        'where',
        'the',
        'mon',
        'tue',
        'wed',
        'thu',
        'fri',
        'sat',
        'sun',
        'monday',
        'tuesday',
        'wednesday',
        'thursday',
        'friday',
        'saturday',
        'sunday',
        'release',
        'released',
        'dance',
        'dancing',
        'need',
        'i need',
        'i would',
        ' i will',
        'find',
        'the list',
        'get some',
        'af',
        '1st',
        '2nd',
        '3rd',
        'tongue',
        'zero',
        'one',
        'two',
        'three',
        'four',
        'five',
        'six',
        'seven',
        'eight',
        'nine',
        'spotify',
        'app',
        'almond',
        'genre',
        'play dj',
        'stone',
        'sound tracks',
        'hi',
        'hey',
        'tweet',
        'all music',
        'hello',
        'preference',
        'top tracks',
        'all the good',
        'music i',
        'id',
        'jan',
        'feb',
        'mar',
        'apr',
        'may',
        'jun',
        'jul',
        'aug',
        'sep',
        'oct',
        'nov',
        'dec',
        'from yesterday',
        'show tunes',
        'tag',
        'ms',
        'all good',
        'greatest hits',
        'good hits',
        'content rating',
        'how long',
        'actor',
        'pg',
        'ratings',
        'rating',
        'rated pg',
        'key',
        'the nice',
        'keyword',
        'keywords',
        'subtitle',
        'subtitles',
        'i want that',
        'shorter',
        'duration',
        'num',
        'hope',
        'rm',
        'michelin',
        'michelin star',
        'michelin stars',
        'reservations',
        'zip code',
        'zipcode',
        'smoke',
        'smoking',
        'luxury',
        'bar',
        'bars',
        'kitchen',
        'cafe',
        'cafes',
        'coffee',
        'where i am',
        'email',
        'motion sensor',
        'temperature sensor',
        'publication date',
        'home address',
        'home assistant',
    ]
)

BANNED_REGEXES = [
    re.compile(r'\d (star|rating)'),
    re.compile(r'\dth'),
    re.compile(r'a \d'),
    re.compile(r'\d (hour|min|sec|minute|second|day|month|year)s?'),
    re.compile(r'this (hour|min|sec|minute|second|day|month|year)s?'),
]


def is_banned(word):
    return word in BANNED_PHRASES or any([regex.match(word) for regex in BANNED_REGEXES])


def normalize_text(text):
    text = unicodedata.normalize('NFD', text).lower()
    text = re.sub('\s\s+', ' ', text)
    return text


def has_overlap(start, end, used_aliases):
    for alias in used_aliases:
        alias_start, alias_end = alias[1], alias[2]
        if start < alias_end and end > alias_start:
            return True
    return False



import random

from ...data_utils.progbar import progress_bar


def parse_argv(parser):
    parser.add_argument('input', type=str, help='The path to the input.')
    parser.add_argument('output1', type=str, help='The path to the output train file.')
    parser.add_argument('output2', type=str, help='The path to the output dev file.')
    parser.add_argument('--output1_ratio', type=float, required=True, help='The ratio of input examples that go to output1')
    parser.add_argument('--seed', default=123, type=int, help='Random seed.')


def main(args):
    random.seed(args.seed)

    with open(args.input, 'r') as input_file, open(args.output1, 'w') as output_file1, open(args.output2, 'w') as output_file2:
        for line in progress_bar(input_file):
            r = random.random()
            if r < args.output1_ratio:
                output_file1.write(line)
            else:
                output_file2.write(line)



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
from collections import OrderedDict

import torch

from .data_utils.progbar import progress_bar
from .metrics import compute_metrics
from .models import TransformerForTokenClassification
from .util import GenerationOutput


def generate_with_model(
    model,
    data_iterator,
    numericalizer,
    task,
    args,
    output_predictions_only=False,
    output_confidence_features=False,
    original_order=None,
    confidence_estimators=None,
    disable_progbar=True,
):

    if isinstance(model, TransformerForTokenClassification):
        return generate_with_classification_model(
            model, data_iterator, numericalizer, task, original_order=original_order, disable_progbar=disable_progbar
        )
    else:
        return generate_with_seq2seq_model(
            model,
            data_iterator,
            numericalizer,
            task,
            args,
            output_predictions_only=output_predictions_only,
            output_confidence_features=output_confidence_features,
            original_order=original_order,
            confidence_estimators=confidence_estimators,
            disable_progbar=disable_progbar,
        )


def generate_with_seq2seq_model(
    model,
    data_iterator,
    numericalizer,
    task,
    args,
    output_predictions_only=False,
    output_confidence_features=False,
    original_order=None,
    confidence_estimators=None,
    disable_progbar=True,
) -> GenerationOutput:
    """
    Inputs:
        original_order: List of indices. If provided, we will sort the results according to this order
        confidence_estimator: if provided, will use it to calculate and output confidence scores
    Outputs: predictions if `output_predictions_only` == True, (loss, predictions, answers, contexts) otherwise
        loss
        predictions: a List of Lists of strings
        answers
        contexts
    """
    output_confidence_scores = confidence_estimators is not None
    predictions = []
    confidence_features = []
    example_ids = []
    answers = []
    contexts = []

    for batch in progress_bar(data_iterator, desc='Generating', disable=disable_progbar):
        batch_size = len(batch.example_id)
        batch_prediction = [[] for _ in range(batch_size)]
        batch_confidence_features = [[] for _ in range(batch_size)]
        batch_example_ids = batch.example_id

        example_ids += batch_example_ids
        if not output_predictions_only:
            batch_answer = numericalizer.reverse(batch.answer.value.data, 'answer')
            batch_answer = [
                task.postprocess_prediction(batch_example_ids[i], batch_answer[i]) for i in range(len(batch_answer))
            ]
            answers += batch_answer
            batch_context = numericalizer.reverse(batch.context.value.data, 'context')
            contexts += batch_context
        elif output_confidence_features:
            # need gold answer for confidence estimation
            batch_answer = numericalizer.reverse(batch.answer.value.data, 'answer')
            answers += batch_answer

        for hyperparameter_idx in range(len(args.temperature)):
            generated = model.generate(
                batch,
                max_output_length=args.max_output_length,
                num_outputs=args.num_outputs[hyperparameter_idx],
                temperature=args.temperature[hyperparameter_idx] if args.temperature[hyperparameter_idx] > 0 else 1.0,
                repetition_penalty=args.repetition_penalty[hyperparameter_idx],
                top_k=args.top_k[hyperparameter_idx],
                top_p=args.top_p[hyperparameter_idx],
                num_beams=args.num_beams[hyperparameter_idx],
                num_beam_groups=args.num_beam_groups[hyperparameter_idx],
                diversity_penalty=args.diversity_penalty[hyperparameter_idx],
                no_repeat_ngram_size=args.no_repeat_ngram_size[hyperparameter_idx],
                do_sample=args.temperature[hyperparameter_idx] != 0,  # if temperature==0, we do not sample
            )
            partial_batch_prediction_ids = generated.sequences
            cross_attentions = getattr(generated, 'cross_attentions', None)

            if cross_attentions is not None:
                # stack tensors to shape (max_output_length, num_layers, batch_size, num_heads, 1, max_input_length)
                cross_attentions = torch.stack(([torch.stack(tuple) for tuple in cross_attentions]))

                # reshape to (num_layers, batch_size, num_heads, max_output_length, max_input_length)
                cross_attentions = cross_attentions.squeeze(4)
                cross_attentions = cross_attentions.permute(1, 2, 3, 0, 4).contiguous()

                # choose only last layer attentions
                # TODO: get penultimate layer of attention vectors instead
                cross_attentions = cross_attentions[-1, ...]

                # postprocess prediction ids
                kwargs = {'numericalizer': numericalizer, 'cross_attentions': cross_attentions}
                partial_batch_prediction_ids = task.batch_postprocess_prediction_ids(
                    batch_example_ids, batch.context.value.data, partial_batch_prediction_ids, **kwargs
                )

            if output_confidence_features or output_confidence_scores:
                partial_batch_confidence_features = model.confidence_features(
                    batch=batch, predictions=partial_batch_prediction_ids, mc_dropout_num=args.mc_dropout_num
                )

            partial_batch_prediction = numericalizer.reverse(partial_batch_prediction_ids, 'answer')

            def get_example_index(i):
                return (i // args.num_outputs[hyperparameter_idx]) % batch_size

            # post-process predictions
            for i in range(len(partial_batch_prediction)):
                partial_batch_prediction[i] = task.postprocess_prediction(
                    batch_example_ids[get_example_index(i)], partial_batch_prediction[i]
                )

            # put them into the right array
            for i in range(len(partial_batch_prediction)):
                batch_prediction[get_example_index(i)].append(partial_batch_prediction[i])
                if output_confidence_features or output_confidence_scores:
                    batch_confidence_features[get_example_index(i)].append(partial_batch_confidence_features[i])

        predictions += batch_prediction
        confidence_features += batch_confidence_features

    if original_order is not None:
        # sort back to the original order
        original_order, example_ids, predictions, answers, contexts, confidence_features = [
            list(a)
            for a in tuple(
                zip(*sorted(list(zip(original_order, example_ids, predictions, answers, contexts, confidence_features))))
            )
        ]

    # TODO calculate and return loss
    loss = None
    output = GenerationOutput(loss=loss)

    if output_predictions_only:
        output.predictions = predictions
    else:
        output.example_ids, output.predictions, output.answers, output.contexts = example_ids, predictions, answers, contexts
    if output_confidence_features:
        output.confidence_features = confidence_features
        if args.override_confidence_labels:
            for i, example in enumerate(confidence_features):
                for confidence in example:
                    confidence.label = answers[i] == args.override_confidence_labels
    if output_confidence_scores:
        output.confidence_scores = []
        for estimator in confidence_estimators:
            confidence_scores = estimator.estimate(confidence_features)
            output.confidence_scores.append(confidence_scores)

    return output


def generate_with_classification_model(
    model, data_iterator, numericalizer, task, original_order=None, disable_progbar=True
) -> GenerationOutput:
    all_example_ids = []
    all_answers = []
    all_contexts = []
    all_predictions = []

    for batch in progress_bar(data_iterator, desc='Generating', disable=disable_progbar):
        batch_example_ids = batch.example_id

        batch_context = numericalizer.reverse(batch.context.value.data, 'context')

        all_example_ids += batch_example_ids

        output = model(input_ids=batch.context.value, attention_mask=(batch.context.value != numericalizer.pad_id))

        labels = batch.answer.value.tolist()

        logits = output.logits
        predictions = torch.argmax(logits, dim=2).tolist()

        # Remove ignored index (special tokens)
        processed_preds = []
        processed_labels = []
        for pred, label in zip(predictions, labels):
            preds_list = []
            labels_list = []
            for p_, l_ in zip(pred, label):
                if l_ == numericalizer.answer_pad_id:
                    continue
                preds_list.append(task.id2label[p_])
                labels_list.append(task.id2label[l_])

            processed_preds.append([" ".join(preds_list)])
            processed_labels.append(" ".join(labels_list))

        all_contexts += batch_context
        all_answers += processed_labels
        all_predictions += processed_preds

    if original_order is not None:
        # sort back to the original order
        original_order, all_example_ids, all_predictions, all_answers, all_contexts = [
            list(a)
            for a in tuple(
                zip(*sorted(list(zip(original_order, all_example_ids, all_predictions, all_answers, all_contexts))))
            )
        ]

    # TODO calculate and return loss
    loss = None
    output = GenerationOutput(
        loss=loss, example_ids=all_example_ids, contexts=all_contexts, answers=all_answers, predictions=all_predictions
    )

    return output


def calculate_and_reduce_metrics(predictions, answers, metrics_to_compute, reduce_metrics):
    metrics = OrderedDict()
    for i in range(len(predictions[0])):
        partial_metrics, _ = compute_metrics([p[i] for p in predictions], answers, metrics_to_compute)
        for k, v in partial_metrics.items():
            if reduce_metrics == 'max':
                metrics[k] = max(metrics.get(k, 0), v)
            else:
                raise ValueError('Invalid reduce_metrics argument')
    return metrics


def print_results(keys, values, num_print=1):
    print()
    start = 0
    end = start + num_print
    values = [val[start:end] for val in values]
    for ex_idx in range(len(values[0])):
        for key_idx, key in enumerate(keys):
            value = values[key_idx][ex_idx]
            v = value[0] if isinstance(value, list) else value
            print(f'{key:>11}: {repr(v)}')
        print()
    sys.stdout.flush()


def validate(task, val_iter, model, numericalizer, args, num_print=10):
    with torch.no_grad():
        model.eval()
        if isinstance(model, torch.nn.DataParallel):
            # get rid of the DataParallel wrapper
            model = model.module

        names = ['beam search', 'answer', 'context']

        output = generate_with_model(model, val_iter, numericalizer, task, args)

        metrics = calculate_and_reduce_metrics(output.predictions, output.answers, task.metrics, args.reduce_metrics)
        results = [output.predictions, output.answers, output.contexts]
        print_results(names, results, num_print=num_print)

        return output, metrics



#
# Copyright (c) 2020, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import shutil

import torch

from . import models
from .calibrate import ConfidenceEstimator
from .util import load_config_json

logger = logging.getLogger(__name__)


def parse_argv(parser):
    parser.add_argument('--path', required=True, help='the model training directory to export')
    parser.add_argument('--embeddings', default='.embeddings/', type=str, help='where to load embeddings from')
    parser.add_argument(
        '--checkpoint_name', default='best.pth', help='Checkpoint file to use (relative to --path, defaults to best.pth)'
    )
    parser.add_argument('-o', '--output', required=True, help='the directory where to export into')


def main(args):
    os.makedirs(args.output, exist_ok=True)
    load_config_json(args)

    # load everything - this will ensure that we initialize the numericalizer correctly
    Model = getattr(models, args.model)
    model, _ = Model.load(
        args.path,
        model_checkpoint_file=args.checkpoint_name,
        args=args,
        device=torch.device('cpu'),
        tasks=[],
    )

    # save the numericalizer to the target directory
    # this will copy over all the necessary vocabulary and config files that the numericalizer needs
    model.numericalizer.save(args.output)

    # now copy over the config.json, checkpoint file, and calibrator files (if any)
    for fn in ['config.json', args.checkpoint_name] + [
        fn for fn in os.listdir(args.path) if ConfidenceEstimator.is_estimator(fn)
    ]:
        src = os.path.join(args.path, fn)
        dst = os.path.join(args.output, fn)
        shutil.copyfile(src, dst)

    logger.info(f'Successfully exported model from {args.path} to {args.output}')



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os

import numpy as np
import torch
import ujson
from bootleg.end2end.bootleg_annotator import BootlegAnnotator
from bootleg.end2end.extract_mentions import extract_mentions
from bootleg.run import run_model
from bootleg.utils.parser.parser_utils import parse_boot_and_emm_args

from .database_utils import is_banned

logger = logging.getLogger(__name__)


def reverse_bisect_left(a, x, lo=None, hi=None):
    """Find item x in list a, and keep it reverse-sorted assuming a
    is reverse-sorted.
    """
    if lo is None:
        lo = 0
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if x > a[mid]:
            hi = mid
        else:
            lo = mid + 1
    return lo


def bootleg_process_examples(ex, bootleg_annotator, args, label, task):
    line = {}
    line['sentence'] = getattr(ex, task.utterance_field())

    assert len(label) == 7
    line['aliases'], line['spans'], line['cands'] = label['aliases'], label['spans'], label['cands']
    line['cand_probs'] = list(map(lambda item: list(item), label['cand_probs']))
    tokens_type_ids, tokens_type_probs = bootleg_annotator.bootleg.collect_features_per_line(
        line, args.bootleg_prob_threshold, getattr(task, 'TTtype2qid', None)
    )

    if task.utterance_field() == 'question':
        for i in range(len(tokens_type_ids)):
            ex.question_feature[i].type_id = tokens_type_ids[i]
            ex.question_feature[i].type_prob = tokens_type_probs[i]

    else:
        for i in range(len(tokens_type_ids)):
            ex.context_feature[i].type_id = tokens_type_ids[i]
            ex.context_feature[i].type_prob = tokens_type_probs[i]

    ex.question_plus_types = task.add_type_tokens(ex.question, ex.question_feature, args.add_types_to_text)
    ex.context_plus_types = task.add_type_tokens(ex.context, ex.context_feature, args.add_types_to_text)

    return ex


def extract_features_with_annotator(examples, bootleg_annotator, args, task):
    with torch.no_grad():
        bootleg_inputs = []
        for ex in examples:
            bootleg_inputs.append(getattr(ex, task.utterance_field()))

        bootleg_labels = bootleg_annotator.label_mentions(bootleg_inputs)

        keys = tuple(bootleg_labels.keys())
        values = list(bootleg_labels.values())
        values_unpacked = list(zip(*values))

        bootleg_labels_unpacked = [dict(zip(keys, values)) for values in values_unpacked]

        for i in range(len(examples)):
            ex = examples[i]
            label = bootleg_labels_unpacked[i]
            examples[i] = bootleg_process_examples(ex, bootleg_annotator, args, label, task)


def init_bootleg_annotator(args, device):
    # instantiate a bootleg object to load config and relevant databases
    bootleg = Bootleg(args)
    bootleg_config = bootleg.create_config(bootleg.fixed_overrides)

    # instantiate the annotator class. we use annotator only in server mode
    # for training we use bootleg functions which preprocess and cache data using multiprocessing, and batching to speed up NED
    bootleg_annotator = BootlegAnnotator(
        config=bootleg_config,
        device='cpu' if device.type == 'cpu' else 'cuda',
        min_alias_len=args.min_entity_len,
        max_alias_len=args.max_entity_len,
        cand_map=bootleg.cand_map,
        threshold=args.bootleg_prob_threshold,
        model_name=args.bootleg_model,
        verbose=False,
    )
    # collect all outputs now; we will filter later
    bootleg_annotator.set_threshold(0.0)
    setattr(bootleg_annotator, 'bootleg', bootleg)
    return bootleg_annotator


class Bootleg(object):
    def __init__(self, args):
        self.args = args

        self.model_dir = f'{self.args.database_dir}/{self.args.bootleg_model}'
        self.config_path = f'{self.model_dir}/bootleg_config.yaml'
        self.cand_map = f'{self.args.database_dir}/wiki_entity_data/entity_mappings/alias2qids.json'

        self.entity_dir = f'{self.args.database_dir}/wiki_entity_data'
        self.embed_dir = f'{self.args.database_dir}/wiki_entity_data'

        self.almond_domains = args.almond_domains
        self.bootleg_post_process_types = args.bootleg_post_process_types

        with open(f'{self.args.database_dir}/wiki_entity_data/type_mappings/wiki/qid2typenames.json', 'r') as fin:
            self.qid2typenames = ujson.load(fin)
        with open(f'{self.args.database_dir}/wiki_entity_data/type_mappings/wiki/type_vocab_to_wikidataqid.json', 'r') as fin:
            self.type_vocab_to_wikidataqid = ujson.load(fin)
            self.wikidataqid_to_type_vocab = {v: k for k, v in self.type_vocab_to_wikidataqid.items()}
        with open(f'{self.args.database_dir}/es_material/typeqid2id.json', 'r') as fin:
            self.typeqid2id = ujson.load(fin)

        # almond specific
        with open(f'{self.args.database_dir}/es_material/almond_type_mapping_matching.json', 'r') as fin:
            self.almond_type_mapping_match = ujson.load(fin)
        with open(f'{self.args.database_dir}/es_material/almond_type_mapping_inclusion.json', 'r') as fin:
            self.almond_type_mapping_include = ujson.load(fin)

        self.cur_entity_embed_size = 0

        # Mapping between model directory and checkpoint name
        model2checkpoint = {'bootleg_uncased_mini': 'bootleg_wiki.pth', 'bootleg_uncased_super_mini': 'bootleg_wiki.pth'}

        self.ckpt_name, extension = model2checkpoint[self.args.bootleg_model].split('.')
        self.model_ckpt_path = os.path.join(self.model_dir, self.ckpt_name + '.' + extension)

        self.fixed_overrides = [
            # emmental configs
            "--emmental.dataparallel",
            'False',
            "--emmental.log_path",
            self.args.bootleg_output_dir,
            "--emmental.use_exact_log_path",
            'True',
            "--emmental.model_path",
            self.model_ckpt_path,
            # run configs
            "--run_config.dataset_threads",
            str(getattr(self.args, 'bootleg_dataset_threads', 1)),
            "--run_config.dataloader_threads",
            str(getattr(self.args, 'bootleg_dataloader_threads', 1)),
            "--run_config.eval_batch_size",
            str(getattr(self.args, 'bootleg_batch_size', 32)),
            "--run_config.log_level",
            'DEBUG',
            # data configs
            "--data_config.print_examples_prep",
            'False',
            "--data_config.entity_dir",
            self.entity_dir,
            "--data_config.entity_prep_dir",
            "prep",
            "--data_config.emb_dir",
            self.embed_dir,
            "--data_config.alias_cand_map",
            'alias2qids.json',
            "--data_config.word_embedding.cache_dir",
            self.args.embeddings,
            "--data_config.print_examples",
            'False',
        ]

    def create_config(self, overrides):
        config_args = parse_boot_and_emm_args(self.config_path, overrides)
        return config_args

    def create_jsonl(self, input_path, examples, utterance_field):
        # create jsonl file for examples
        jsonl_input_path = input_path.rsplit('.', 1)[0] + '.jsonl'
        with open(jsonl_input_path, 'w') as fout:
            for ex in examples:
                fout.write(ujson.dumps({"sentence": getattr(ex, utterance_field)}) + '\n')

    def extract_mentions(self, input_path):
        jsonl_input_path = input_path.rsplit('.', 1)[0] + '.jsonl'
        jsonl_output_path = input_path.rsplit('.', 1)[0] + '_bootleg.jsonl'
        logger.info('Extracting mentions...')
        extract_mentions(
            in_filepath=jsonl_input_path,
            out_filepath=jsonl_output_path,
            cand_map_file=self.cand_map,
            min_alias_len=self.args.min_entity_len,
            max_alias_len=self.args.max_entity_len,
            num_workers=getattr(self.args, 'bootleg_extract_num_workers', 32),
            verbose=False,
        )

    def pad_values(self, tokens, max_size, pad_id):
        if len(tokens) > max_size:
            tokens = tokens[:max_size]
        else:
            tokens += [pad_id] * (max_size - len(tokens))
        return tokens

    def disambiguate_mentions(self, config_args):
        run_model(self.args.bootleg_dump_mode, config_args)

    def post_process_bootleg_types(self, title, TTtype2qid):
        type_mapping_match = dict()
        type_mapping_include = dict()
        for domain in self.almond_domains:
            type_mapping_match.update(self.almond_type_mapping_match[domain])
            type_mapping_include.update(self.almond_type_mapping_match[domain])

        types = None
        if title in type_mapping_match:
            types = type_mapping_match[title]
        else:
            for key in type_mapping_include.keys():
                if key in title:
                    types = type_mapping_include[key]

        typeqids = None
        if types is not None:
            if isinstance(types, str):
                typeqids = [TTtype2qid[types]]
            elif isinstance(types, (list, tuple)):
                typeqids = [TTtype2qid[type_] for type_ in types]

        return typeqids

    def collect_features_per_line(self, line, threshold, TTtype2qid):

        tokenized = line['sentence'].split(' ')
        tokens_type_ids = [
            [self.args.ned_features_default_val[0]] * self.args.ned_features_size[0] for _ in range(len(tokenized))
        ]
        tokens_type_probs = [
            [self.args.ned_features_default_val[1]] * self.args.ned_features_size[1] for _ in range(len(tokenized))
        ]

        for alias, all_qids, all_probs, span in zip(line['aliases'], line['cands'], line['cand_probs'], line['spans']):
            # filter qids with probability lower than a threshold
            idx = reverse_bisect_left(all_probs, threshold)
            all_qids = all_qids[:idx]
            all_probs = all_probs[:idx]

            type_ids = []
            type_probs = []

            if not is_banned(alias):
                for qid, prob in zip(all_qids, all_probs):
                    # get all type for a qid
                    all_types = []
                    if qid in self.qid2typenames and self.qid2typenames[qid]:
                        # map entity qid to its type titles on wikidata ; then map titles to their wikidata qids
                        for typename in self.qid2typenames[qid]:
                            if typename in self.type_vocab_to_wikidataqid:
                                all_types.append(self.type_vocab_to_wikidataqid[typename])

                    if len(all_types):
                        # update
                        # go through all types
                        for typeqid in all_types:
                            if typeqid in self.typeqid2id:
                                # map wikidata types to thingtalk types
                                if self.bootleg_post_process_types:
                                    # map qid to title
                                    title = self.wikidataqid_to_type_vocab[typeqid]
                                    # process may return multiple types for a single type when it's ambiguous
                                    typeqids = self.post_process_bootleg_types(title, TTtype2qid)
                                    if typeqids is None:
                                        typeqids = [typeqid]

                                else:
                                    typeqids = [typeqid]

                                for typeqid_ in typeqids:
                                    type_id = self.typeqid2id[typeqid_]
                                    type_ids.append(type_id)
                                    type_probs.append(prob)

                padded_type_ids = self.pad_values(
                    type_ids, self.args.ned_features_size[0], self.args.ned_features_default_val[0]
                )
                padded_type_probs = self.pad_values(
                    type_probs, self.args.ned_features_size[1], self.args.ned_features_default_val[1]
                )

                tokens_type_ids[span[0] : span[1]] = [padded_type_ids] * (span[1] - span[0])
                tokens_type_probs[span[0] : span[1]] = [padded_type_probs] * (span[1] - span[0])

        return tokens_type_ids, tokens_type_probs

    def collect_features(self, file_name, subsample, TTtype2qid):

        all_tokens_type_ids = []
        all_tokens_type_probs = []

        threshold = self.args.bootleg_prob_threshold

        with open(f'{self.args.bootleg_output_dir}/{file_name}_bootleg/{self.ckpt_name}/bootleg_labels.jsonl', 'r') as fin:
            for i, line in enumerate(fin):
                if i >= subsample:
                    break
                line = ujson.loads(line)
                tokens_type_ids, tokens_type_probs = self.collect_features_per_line(line, threshold, TTtype2qid)
                all_tokens_type_ids.append(tokens_type_ids)
                all_tokens_type_probs.append(tokens_type_probs)

        if os.path.exists(f'{self.args.bootleg_output_dir}/{file_name}_bootleg/{self.ckpt_name}/bootleg_embs.npy'):
            with open(f'{self.args.bootleg_output_dir}/{file_name}_bootleg/{self.ckpt_name}/bootleg_embs.npy', 'rb') as fin:
                emb_data = np.load(fin)
                self.cur_entity_embed_size += emb_data.shape[0]

        return all_tokens_type_ids, all_tokens_type_probs

    def merge_embeds(self, file_list):
        all_emb_data = []
        for file_name in file_list:
            emb_file = f'{self.args.bootleg_output_dir}/{file_name}_bootleg/{self.ckpt_name}/bootleg_embs.npy'
            with open(emb_file, 'rb') as fin:
                emb_data = np.load(fin)
                all_emb_data.append(emb_data)

        all_emb_data = np.concatenate(all_emb_data, axis=0)

        # add embeddings for the padding and unknown special tokens
        new_emb = np.concatenate([np.zeros([2, all_emb_data.shape[1]], dtype='float'), all_emb_data], axis=0)

        os.makedirs(f'{self.args.bootleg_output_dir}/bootleg/eval/{self.ckpt_name}', exist_ok=True)
        np.save(f'{self.args.bootleg_output_dir}/bootleg/eval/{self.ckpt_name}/ent_embedding.npy', new_emb)



#
# Copyright (c) 2020, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

from .progbar import progress_bar

quoted_pattern_maybe_space = re.compile(r'\"\s?([^"]*?)\s?\"')
quoted_pattern_with_space = re.compile(r'\"\s([^"]*?)\s\"')
device_pattern = re.compile(r'\s@([\w\.]+)\s')
entity_regex = re.compile('<e>.*?</e>')
token_type_regex = re.compile('(.*?) \( (.*?) \)')

ISO_to_LANG = {
    'en': 'English',
    'en-US': 'English',
    'fa': 'Persian',
    'it': 'Italian',
    'zh': 'Chinese',
    'hr': 'Croatian',
    'ja': 'Japanese',
    'ko': 'Korean',
    'ru': 'Russian',
    'es': 'Spanish',
    'sv': 'Swedish',
    'tr': 'Turkish',
    'hi': 'Hindi',
    'fr': 'French',
    'de': 'German',
    'pl': 'Polsih',
    'ar': 'Arabic',
    'vi': 'Vietnamese',
    'ji': 'Yiddish',
    'pt': 'Portuguese',
    'el': 'Greek',
    'he': 'Hebrew',
    'si': 'Sinhala',
    'ta': 'Tamil',
    'fi': 'Finnish',
    'cs': 'Czech',
    'no': 'Norwegian',
    'tl': 'Filipino',
    'da': 'Danish',
}


CJK_RANGES = [
    (ord(u"\u3300"), ord(u"\u33ff")),
    (ord(u"\ufe30"), ord(u"\ufe4f")),  # compatibility ideographs
    (ord(u"\uf900"), ord(u"\ufaff")),
    (ord(u"\U0002F800"), ord(u"\U0002fa1f")),  # compatibility ideographs
    (ord(u'\u3040'), ord(u'\u309f')),  # Japanese Hiragana
    (ord(u"\u30a0"), ord(u"\u30ff")),  # Japanese Katakana
    (ord(u"\u2e80"), ord(u"\u2eff")),  # cjk radicals supplement
    (ord(u"\u4e00"), ord(u"\u9fff")),
    (ord(u"\u3400"), ord(u"\u4dbf")),
    (ord(u"\U00020000"), ord(u"\U0002a6df")),
    (ord(u"\U0002a700"), ord(u"\U0002b73f")),
    (ord(u"\U0002b740"), ord(u"\U0002b81f")),
    (ord(u"\U0002b820"), ord(u"\U0002ceaf")),
]

CJK_ADDONS = [ord(u"\u3001")]


def is_cjk_char(cp):
    return cp in CJK_ADDONS or any([range[0] <= cp <= range[1] for range in CJK_RANGES])


ENTITY_REGEX = re.compile('^[A-Z]+_')


def is_entity(token):
    return ENTITY_REGEX.match(token) is not None


def is_device(token):
    return token[0] == '@'


def is_entity_marker(token):
    return token.startswith('^^')


def process_id(ex):
    # Example instance
    if isinstance(ex.example_id, str):
        id_ = ex.example_id.rsplit('/', 1)
    # NumericalizedExample instance
    else:
        assert isinstance(ex.example_id, list)
        assert len(ex.example_id) == 1
        id_ = ex.example_id[0].rsplit('/', 1)
    id_ = id_[0] if len(id_) == 1 else id_[1]
    # translated
    if id_[0] == 'T':
        id_ = id_[1:]
    return id_


def tokenize_cjk_chars(sentence):
    output = []
    i = 0
    while i < len(sentence):
        output.append(sentence[i])
        if is_cjk_char(ord(sentence[i])) and i + 1 < len(sentence) and sentence[i + 1] != ' ':
            output.append(' ')
        elif not is_cjk_char(ord(sentence[i])) and i + 1 < len(sentence) and is_cjk_char(ord(sentence[i + 1])):
            output.append(' ')
        i += 1

    output = "".join(output)
    output = output.replace('  ', ' ')

    return output


def detokenize_cjk_chars(sentence):
    output = []
    i = 0
    while i < len(sentence):
        output.append(sentence[i])
        # skip space after cjk chars only if followed by another cjk char
        if (
            is_cjk_char(ord(sentence[i]))
            and i + 1 < len(sentence)
            and sentence[i + 1] == ' '
            and i + 2 < len(sentence)
            and is_cjk_char(ord(sentence[i + 2]))
        ):
            i += 2
        else:
            i += 1

    return "".join(output)


def chunk_file(input_src, chunk_files, chunk_size, num_chunks):
    chunk_id = 0
    num_lines_in_chunk = 0
    all_out_files = [open(chunk_files[chunk_id], 'w') for chunk_id in range(num_chunks)]
    with open(input_src, 'r', encoding='utf-8') as in_file:
        for line in in_file:
            all_out_files[chunk_id].write(line)
            num_lines_in_chunk += 1
            if num_lines_in_chunk == chunk_size:
                chunk_id += 1
                num_lines_in_chunk = 0
                if chunk_id == num_chunks:
                    break
    for file in all_out_files:
        file.close()


def create_examples_from_file(args):
    path = args['in_file']
    chunk_size = args['chunk_size']
    dir_name = args['dir_name']
    example_batch_size = args['example_batch_size']
    make_process_example = args['make_process_example']
    kwargs = args['kwargs']

    chunk_examples = []

    batch = []
    last_batch = False
    for i, line in progress_bar(enumerate(open(path, 'r', encoding='utf-8')), desc='Reading dataset'):
        parts = line.strip().split('\t')
        batch.append(parts)
        if len(chunk_examples) + example_batch_size > chunk_size:
            # trim batch
            batch = batch[: chunk_size - len(chunk_examples)]
            last_batch = True
        if len(batch) % example_batch_size != 0 and not last_batch:
            continue

        # TODO remote database lookup is faster when multiple examples are sent in one HTTP request
        # For now we only do one at a time; support processing examples as a batch
        assert len(batch) == 1
        batch = batch[0]
        chunk_examples.append(make_process_example(batch, dir_name, **kwargs))
        batch = []
        if len(chunk_examples) >= chunk_size:
            break

    return chunk_examples



import csv
import os
import random
import re
import sys

from ...data_utils.progbar import progress_bar
from ...util import detokenize

csv.field_size_limit(sys.maxsize)


def is_english(s):
    try:
        s.encode(encoding='utf-8').decode('ascii')
    except UnicodeDecodeError:
        return False
    else:
        return True


def normalize(s):
    # remove quotations
    s = s.replace('``', '')
    s = s.replace('\'\'', '')
    s = s.replace('"', '')
    if s.startswith('\''):
        s = s[1:]
    if s.endswith('\''):
        s = s[:-1]

    s = s.replace('...', '')
    s = s.replace(',', ', ')
    s = re.sub('\s\s+', ' ', s)  # remove duplicate white spaces
    s = s.strip()

    return s


def is_valid(s):
    return (
        'http' not in s
        and s.count('-') <= 4
        and s.count('.') <= 4
        and is_english(s)
        and '_' not in s
        and '%' not in s
        and '/' not in s
        and '[' not in s
        and ']' not in s
        and '*' not in s
        and '\\' not in s
        and 'www' not in s
        and sum(c.isdigit() for c in s) <= 10
        and s.count('(') == s.count(')')
    )


def normalized_levenshtein(s1, s2, mode='character'):
    if mode != 'character' and isinstance(s1, str):
        s1 = s1.split()
        s2 = s2.split()
    if len(s1) < len(s2):
        return normalized_levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = (
                previous_row[j + 1] + 1
            )  # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1] / max(len(s1), len(s2))


def pos_tag_string(sentence: str):
    # load NLTK lazily
    import nltk

    nltk.download('averaged_perceptron_tagger', quiet=True)
    tagged_tokens = nltk.pos_tag(nltk.word_tokenize(sentence))
    tags = [t[1] for t in tagged_tokens]
    return ' '.join(tags).lower()


def parse_argv(parser):
    parser.add_argument('input', type=str, help='The path to the input .tsv file.')
    parser.add_argument('output_dir', type=str, help='The path to the folder to save train.tsv and dev.tsv files.')

    parser.add_argument(
        '--train_ratio', type=float, default=0.95, help='The ratio of input examples that go to the training set'
    )
    parser.add_argument('--seed', default=123, type=int, help='Random seed used for train/dev split.')

    # By default, we swap the columns so that the target of paraphrasing will be a grammatically correct sentence, i.e. written by a human, not an NMT
    parser.add_argument(
        '--first_columns',
        type=int,
        nargs='+',
        default=[2],
        help='The column indices in the input file to put in the first column of the output file',
    )
    parser.add_argument(
        '--second_columns',
        type=int,
        nargs='+',
        default=[1],
        help='The column indices in the input file to put in the second column of the output file',
    )

    parser.add_argument(
        '--min_length',
        type=int,
        default=20,
        help='Minimum number of characters that each phrase should have in order to be included',
    )
    parser.add_argument(
        '--max_length',
        type=int,
        default=200,
        help='Maximum number of characters that each phrase should have in order to be included',
    )
    parser.add_argument('--edit_distance_mode', type=str, default='none', choices=['character', 'word', 'none'])
    parser.add_argument(
        '--min_edit_distance',
        type=float,
        default=0.0001,
        help='We will not include phrase pairs that have a normalized edit distance below this number.',
    )
    parser.add_argument(
        '--plot_edit_distance', action='store_true', help='Save a plot of the normalized edit distance distribution.'
    )
    parser.add_argument('--skip_check', action='store_true', help='Skip validity check.')
    parser.add_argument('--skip_normalization', action='store_true', help='Do not remove quotation marks or detokenize.')
    parser.add_argument('--lower_case', action='store_true', help='Convert everything to lower case.')
    parser.add_argument(
        '--prepend_pos_tags',
        action='store_true',
        help='Prepend the part-of-speech tag of the output to the beginning of the input.',
    )
    parser.add_argument(
        '--sep_token', type=str, default='</s>', help='Token to insert between the part-of-speech tag and the input sentence.'
    )
    parser.add_argument('--max_examples', type=int, default=1e10, help='Maximum number of examples in the output.')


def main(args):
    random.seed(args.seed)

    drop_count = 0
    # number_of_lines = get_number_of_lines(args.input)
    # number_of_lines = get_number_of_lines(args.input)
    output_size = 0
    included_examples = 0
    sum_edit_distance = 0
    all_normalized_edit_distances = []
    with open(args.input, 'r') as input_file, open(os.path.join(args.output_dir, 'train.tsv'), 'w') as train_output_file, open(
        os.path.join(args.output_dir, 'dev.tsv'), 'w'
    ) as dev_output_file:
        train_writer = csv.writer(train_output_file, delimiter='\t')
        dev_writer = csv.writer(dev_output_file, delimiter='\t')
        reader = csv.reader(input_file, delimiter='\t')
        for row in progress_bar(reader, desc='Lines'):
            is_written = False

            # Decide which output file this example should be written to. Note that all the examples generated from this row will go to the same file
            r = random.random()
            if r < args.train_ratio:
                writer = train_writer
            else:
                writer = dev_writer

            for first_column in args.first_columns:
                if first_column >= len(row):
                    continue
                for second_column in args.second_columns:
                    if second_column >= len(row) or first_column == second_column:
                        continue
                    first = row[first_column]  # input sequence
                    second = row[second_column]  # output_sequence
                    # print('first = ', first)
                    # print('second = ', second)
                    if not args.skip_check and (
                        len(first) < args.min_length
                        or len(second) < args.min_length
                        or len(first) > args.max_length
                        or len(second) > args.max_length
                        or not is_valid(first)
                        or not is_valid(second)
                    ):
                        drop_count += 1
                        continue
                    if not args.skip_normalization:
                        first = normalize(detokenize(first))
                        second = normalize(detokenize(second))
                    first = first.strip()
                    second = second.strip()
                    if args.lower_case:
                        first = first.lower()
                        second = second.lower()
                    if first == '' or second == '' or first.lower() == second.lower():
                        drop_count += 1
                        continue
                    if args.edit_distance_mode != 'none':
                        normalized_edit_distance = normalized_levenshtein(
                            first.lower(), second.lower(), mode=args.edit_distance_mode
                        )
                        # print('normalized_edit_distance = ', normalized_edit_distance)
                        if normalized_edit_distance < args.min_edit_distance:
                            drop_count += 1
                            continue
                        all_normalized_edit_distances.append(normalized_edit_distance)
                        sum_edit_distance += normalized_edit_distance
                    if args.prepend_pos_tags:
                        first = pos_tag_string(second) + ' ' + args.sep_token + ' ' + first
                    writer.writerow([first, second])
                    output_size += 1
                    is_written = True
            if is_written:
                included_examples += 1
                if included_examples >= args.max_examples:
                    break

    print('Dropped', drop_count, 'examples')
    print('Average normalized edit distance between pairs is ', sum_edit_distance / output_size)
    if args.edit_distance_mode != 'none' and args.plot_edit_distance:
        import matplotlib.pyplot as plt

        _, _, _ = plt.hist(all_normalized_edit_distances, 20)
        plt.savefig(os.path.join(args.output_dir, 'edit_distance_plot.png'))



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from . import generic_dataset


class BaseTask(object):
    """
    Base class for all tasks.

    Includes all the code to handle generic tasks

    """

    def __init__(self, name, args):
        self.name = name
        # special task-specific tokens that should not be subword tokenized
        self.special_tokens = set()
        self.override_context = args.override_context
        self.override_question = args.override_question

    @property
    def default_question(self):
        return ''

    @property
    def default_context(self):
        return ''

    @property
    def utterance_field(self):
        return NotImplementedError

    def get_splits(self, root, **kwargs):
        """
        Load the train, test, eval datasets for this task

        :param field: the text.Field to use for tokenization, preprocessing and vocabulary construction
        :param root: the base directory where data is stored
        :param kwargs: other arguments to pass to the Dataset
        :return: a list of text.Dataset
        """
        return generic_dataset.JSON.splits(root=root, name=self.name, **kwargs)

    def batch_postprocess_prediction_ids(self, batch_example_ids, batch_src_ids, batch_tgt_ids, **kwargs):
        return batch_tgt_ids

    def postprocess_prediction(self, example_id, prediction):
        return prediction

    def preprocess_field(self, sentence, field_name=None, answer=None):
        if self.override_context is not None and field_name == 'context':
            return self.override_context
        if self.override_question is not None and field_name == 'question':
            return self.override_question
        return sentence, [], sentence

    @property
    def metrics(self):
        """
        What metrics to evaluate this task on.

        This property must return a non-empty list.
        The first entry in the list will be the metric to use to compute the decascore.

        :return: a list of metric names
        """
        return ['em', 'nem', 'nf1']



import csv
import re

from ...data_utils.progbar import progress_bar
from ...util import lower_case, remove_thingtalk_quotes, tokenize

special_token_pattern = re.compile("(^|(?<= ))" + "[A-Z]+_[0-9]" + "($|(?= ))")


def find_special_tokens(s: str):
    return list(sorted([a.group(0) for a in special_token_pattern.finditer(s)]))


def is_subset(set1, set2):
    """
    Returns True if set1 is a subset of or equal to set2
    """
    return all([e in set2 for e in set1])


def passes_heuristic_checks(row, args, old_query=None):
    if 'QUOTED_STRING' in row[args.utterance_column] or (old_query is not None and 'QUOTED_STRING' in old_query):
        # remove quoted examples
        return False
    if old_query is not None:
        # check that all the special tokens in utterance after paraphrasing are the same as before
        if find_special_tokens(old_query) != find_special_tokens(row[args.utterance_column]):
            return False
    all_input_columns = ' '.join([row[c] for c in args.input_columns])
    input_special_tokens = set(find_special_tokens(all_input_columns))
    output_special_tokens = set(find_special_tokens(row[args.thingtalk_column]))
    if not is_subset(output_special_tokens, input_special_tokens):
        return False
    _, quote_values = remove_thingtalk_quotes(row[args.thingtalk_column])
    if quote_values is None:
        return False  # ThingTalk is not syntactic, so remove this example
    for q in quote_values:
        if q not in all_input_columns:
            return False
    return True


def parse_argv(parser):
    parser.add_argument('input', type=str, help='The path to the input file.')
    parser.add_argument('output', type=str, help='The path to the output file.')
    parser.add_argument('--query_file', type=str, help='The path to the file containing new queries.')
    parser.add_argument(
        '--thrown_away',
        type=str,
        default=None,
        help='The path to the output file that will contain inputs that were removed because of `--transformation`.',
    )
    parser.add_argument(
        '--thingtalk_gold_file', type=str, help='The path to the file containing the dataset with a correct thingtalk column.'
    )
    parser.add_argument(
        '--transformation',
        type=str,
        choices=[
            'remove_thingtalk_quotes',
            'replace_queries',
            'remove_wrong_thingtalk',
            'get_wrong_thingtalk',
            'merge_input_file_with_query_file',
            'none',
        ],
        default='none',
        help='The type of transformation to apply.',
    )
    parser.add_argument(
        '--remove_duplicates',
        action='store_true',
        help='Remove duplicate natural utterances. Note that this also removes cases where a natural utterance has multiple ThingTalk codes.',
    )

    # These arguments are effective only if --transformation=get_wrong_thingtalk
    parser.add_argument(
        '--remove_with_heuristics',
        action='store_true',
        help='Remove examples if the values inside quotations in ThingTalk have changed or special words like NUMBER_0 cannot be found in TT anymore.',
    )
    parser.add_argument(
        '--replace_with_gold',
        action='store_true',
        help='Instead of the original ThingTalk, output what the parser said is gold.',
    )
    parser.add_argument(
        '--task',
        type=str,
        required=True,
        choices=['almond', 'almond_dialogue_nlu', 'almond_dialogue_nlu_agent'],
        help='Specifies the meaning of columns in the input file and the ones that should go to the output',
    )

    # parser.add_argument('--output_columns', type=int, nargs='+', default=None,
    #                     help='The columns to write to output. By default, we output all columns.')
    # parser.add_argument('--id_column', type=int, default=0,
    #                     help='The column index in the input file that contains the unique id')
    # parser.add_argument('--utterance_column', type=int, default=1,
    #                     help='The column index in the input file that contains the natural utterance')
    # parser.add_argument('--thingtalk_column', type=int, default=2,
    #                     help='The column index in the input file that contains the ThingTalk code.')
    # parser.add_argument('--no_duplication_columns', type=int, nargs='+', default=None,
    #                     help='The columns indices in the input file that determine whether two rows are duplicates of each other or not.')


def main(args):
    if args.task == 'almond':
        args.id_column = 0
        args.utterance_column = 1
        args.thingtalk_column = 2
        args.output_columns = [0, 1, 2]
        args.no_duplication_columns = [1]
        args.input_columns = [1]

    elif args.task == 'almond_dialogue_nlu' or 'almond_dialogue_nlu_agent':
        args.id_column = 0
        # column 1 is ontext (ThingTalk)
        args.utterance_column = 2
        args.thingtalk_column = 3
        args.output_columns = [0, 1, 2, 3]
        args.no_duplication_columns = [1, 2]
        args.input_columns = [1, 2]

    if args.output_columns is None:
        # if args.transformation == 'remove_wrong_thingtalk':
        # args.output_columns = [0, 1, 2, 3, 4] # we add original utterance and ThingTalk as well
        args.output_columns = [0, 1, 2]
    if args.remove_duplicates and args.no_duplication_columns is None:
        raise ValueError('You should specify columns that define duplication')

    with open(args.input, 'r') as input_file, open(args.output, 'w') as output_file:
        reader = csv.reader(input_file, delimiter='\t')
        if args.transformation in ['replace_queries', 'merge_input_file_with_query_file']:
            new_queries = []  # list of lists
            query_file = open(args.query_file, 'r')
            for line in query_file:
                queries = line.split('\t')[1:]  # 0 is example id
                new_queries.append([lower_case(tokenize(q.strip())) for q in queries])
        if args.transformation in ['remove_wrong_thingtalk', 'get_wrong_thingtalk']:
            gold_thingtalks = []
            thingtalk_gold_file_reader = csv.reader(open(args.thingtalk_gold_file, 'r'), delimiter='\t')
            for q in thingtalk_gold_file_reader:
                gold_thingtalks.append(q[1].strip())
            gold_thingtalk_count = 0

        duplicate_count = 0
        heuristic_count = 0
        written_count = 0
        if args.remove_duplicates:
            seen_examples = set()
        all_thrown_away_rows = []
        for row_idx, row in enumerate(progress_bar(reader, desc='Lines')):
            old_query = None
            output_rows = []
            thrown_away_rows = []
            if args.transformation == 'remove_thingtalk_quotes':
                row[args.thingtalk_column], _ = remove_thingtalk_quotes(row[args.thingtalk_column])
            if args.transformation == 'remove_wrong_thingtalk':
                if row[args.thingtalk_column] == gold_thingtalks[gold_thingtalk_count]:
                    output_rows = [row]
                else:
                    output_rows = []
                    thrown_away_rows = [row]
                gold_thingtalk_count += 1
            elif args.transformation == 'get_wrong_thingtalk':
                if row[args.thingtalk_column] != gold_thingtalks[gold_thingtalk_count]:
                    if args.replace_with_gold:
                        row[args.thingtalk_column] = gold_thingtalks[gold_thingtalk_count]
                    output_rows = [row]
                else:
                    output_rows = []
                gold_thingtalk_count += 1
            elif args.transformation == 'merge_input_file_with_query_file':
                output_rows.append(row)
                for new_query in new_queries[row_idx]:
                    row = row.copy()
                    row[args.utterance_column] = new_query
                    output_rows.append(row)
            elif args.transformation == 'replace_queries':
                old_query = row[args.utterance_column]
                for idx, new_query in enumerate(new_queries[row_idx]):
                    copy_row = row.copy()
                    copy_row[args.utterance_column] = new_query
                    copy_row[args.id_column] = 'A' + copy_row[args.id_column] + '-' + str(idx)  # add 'A' for auto-paraphrasing
                    output_rows.append(copy_row)
            else:
                assert args.transformation == 'none'
                # do basic clean-up because old generation code missed these
                row[args.utterance_column] = row[args.utterance_column].replace('<pad>', '')
                row[args.utterance_column] = re.sub('\s\s+', ' ', row[args.utterance_column])
                row[args.utterance_column] = row[args.utterance_column].strip()
                output_rows = [row]

            for o in output_rows:
                output_row = ""
                if args.remove_with_heuristics:
                    if not passes_heuristic_checks(o, args, old_query=old_query):
                        heuristic_count += 1
                        continue
                if args.remove_duplicates:
                    normalized_example = re.sub('\s+', '', ''.join([o[i] for i in args.no_duplication_columns]))
                    # print(normalized_example)
                    if normalized_example in seen_examples:
                        duplicate_count += 1
                        continue
                    else:
                        seen_examples.add(normalized_example)
                written_count += 1
                for i, column in enumerate(args.output_columns):
                    output_row += o[column]
                    if i < len(args.output_columns) - 1:
                        output_row += '\t'
                output_file.write(output_row + '\n')
            for o in thrown_away_rows:
                if not args.remove_with_heuristics or (
                    args.remove_with_heuristics and passes_heuristic_checks(o, args, old_query=old_query)
                ):
                    all_thrown_away_rows.append(o)

        if args.thrown_away is not None:
            # write the thrown away exampels into a file
            with open(args.thrown_away, 'w') as output_file:
                for o in all_thrown_away_rows:
                    output_row = ""
                    for i, column in enumerate(args.output_columns):
                        output_row += o[column]
                        if i < len(args.output_columns) - 1:
                            output_row += '\t'
                    output_file.write(output_row + '\n')

        print('Removed %d duplicate rows' % duplicate_count)
        print('Removed %d rows because the thingtalk quote did not satisfy heuristic rules' % heuristic_count)
        print('Output size is %d rows' % written_count)



#
# Copyright (c) 2018, Salesforce, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import os
import re
import string
from contextlib import closing
from multiprocessing import Pool, cpu_count
from subprocess import PIPE, Popen
from typing import Iterable

import numpy as np
from pyrouge import Rouge155
from sacrebleu import corpus_bleu
from seqeval import metrics as seq_metrics
from seqeval import scheme as seq_scheme

from .tasks.generic_dataset import Query
from .util import requote_program


def to_lf(s, table):
    aggs = [y.lower() for y in Query.agg_ops]
    agg_to_idx = {x: i for i, x in enumerate(aggs)}
    conditionals = [y.lower() for y in Query.cond_ops]
    headers_unsorted = [(y.lower(), i) for i, y in enumerate(table['header'])]
    headers = [(y.lower(), i) for i, y in enumerate(table['header'])]
    headers.sort(reverse=True, key=lambda x: len(x[0]))
    condition_s, conds = None, []
    if 'where' in s:
        s, condition_s = s.split('where', 1)

    s = ' '.join(s.split()[1:-2])
    sel, agg = None, 0
    for col, idx in headers:
        if col == s:
            sel = idx
    if sel is None:
        s = s.split()
        agg = agg_to_idx[s[0]]
        s = ' '.join(s[1:])
        for col, idx in headers:
            if col == s:
                sel = idx

    full_conditions = []
    if condition_s is not None:

        condition_s = ' ' + condition_s + ' '
        for idx, col in enumerate(headers):
            condition_s = condition_s.replace(' ' + col[0] + ' ', ' Col{} '.format(col[1]))
        condition_s = condition_s.strip()

        for idx, col in enumerate(conditionals):
            new_s = []
            for t in condition_s.split():
                if t == col:
                    new_s.append('Cond{}'.format(idx))
                else:
                    new_s.append(t)
            condition_s = ' '.join(new_s)
        s = condition_s
        conds = re.split('(Col\d+ Cond\d+)', s)
        if len(conds) == 0:
            conds = [s]
        conds = [x for x in conds if len(x.strip()) > 0]
        full_conditions = []
        for i, x in enumerate(conds):
            if i % 2 == 0:
                x = x.split()
                col_num = int(x[0].replace('Col', ''))
                opp_num = int(x[1].replace('Cond', ''))
                full_conditions.append([col_num, opp_num])
            else:
                x = x.split()
                if x[-1] == 'and':
                    x = x[:-1]
                x = ' '.join(x)
                if 'Col' in x:
                    new_x = []
                    for t in x.split():
                        if 'Col' in t:
                            idx = int(t.replace('Col', ''))
                            t = headers_unsorted[idx][0]
                        new_x.append(t)
                    x = new_x
                    x = ' '.join(x)
                if 'Cond' in x:
                    new_x = []
                    for t in x.split():
                        if 'Cond' in t:
                            idx = int(t.replace('Cond', ''))
                            t = conditionals[idx]
                        new_x.append(t)
                    x = new_x
                    x = ' '.join(x)
                full_conditions[-1].append(x)
    logical_form = {'sel': sel, 'conds': full_conditions, 'agg': agg}
    return logical_form


def computeLFEM(greedy, answer):
    answer = [x[0] for x in answer]
    count = 0
    correct = 0
    text_answers = []
    for idx, (g, ex) in enumerate(zip(greedy, answer)):
        count += 1
        text_answers.append([ex['answer'].lower()])
        try:
            lf = to_lf(g, ex['table'])
            gt = ex['sql']
            conds = gt['conds']
            lower_conds = []
            for c in conds:
                lc = c
                lc[2] = str(lc[2]).lower()
                lower_conds.append(lc)
            gt['conds'] = lower_conds
            correct += lf == gt
        except BaseException:
            continue
    return correct / count * 100, text_answers


def score(answer, gold):
    if len(gold) > 0:
        gold = set.union(*[simplify(g) for g in gold])
    answer = simplify(answer)
    tp, tn, sys_pos, real_pos = 0, 0, 0, 0
    if answer == gold:
        if not ('unanswerable' in gold and len(gold) == 1):
            tp += 1
        else:
            tn += 1
    if not ('unanswerable' in answer and len(answer) == 1):
        sys_pos += 1
    if not ('unanswerable' in gold and len(gold) == 1):
        real_pos += 1
    return np.array([tp, tn, sys_pos, real_pos])


def simplify(answer):
    simplified = answer.strip().lower().split()
    simplified = (''.join(c for c in t if c not in string.punctuation) for t in simplified)
    return set(simplified) - {'the', 'a', 'an', 'and', ''}


# http://nlp.cs.washington.edu/zeroshot/evaluate.py
def computeCF1(greedy, answer):
    scores = np.zeros(4)
    for g, a in zip(greedy, answer):
        scores += score(g, a)
    tp, tn, sys_pos, real_pos = scores.tolist()
    if tp == 0:
        p = r = f = 0.0
    else:
        p = tp / float(sys_pos)
        r = tp / float(real_pos)
        f = 2 * p * r / (p + r)

    return f * 100, p * 100, r * 100


def normalize_text(s):
    """Lower text and remove punctuation, articles and extra whitespace."""

    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, ground_truth):
    prediction_tokens = prediction.split()
    ground_truth_tokens = ground_truth.split()
    common = collections.Counter(prediction_tokens) & collections.Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def exact_match(prediction, ground_truth):
    return prediction == ground_truth


def partial_exact_match(prediction, ground_truth):
    prediction = prediction.split()
    ground_truth = ground_truth.split()
    is_correct_token = [p == g for p, g in zip(prediction, ground_truth)]
    partial_score = sum(is_correct_token) / len(is_correct_token)
    return partial_score


def structure_match(prediction, ground_truth):
    return requote_program(prediction) == requote_program(ground_truth)


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    scores_for_ground_truths = []
    for idx, ground_truth in enumerate(ground_truths):
        score = metric_fn(prediction, ground_truth)
        scores_for_ground_truths.append(score)
    return max(scores_for_ground_truths)


def computeF1(outputs, targets):
    outs = [metric_max_over_ground_truths(f1_score, o, t) for o, t in zip(outputs, targets)]
    return sum(outs) / len(outputs) * 100


def computeEM(outputs, targets):
    outs = [metric_max_over_ground_truths(exact_match, o, t) for o, t in zip(outputs, targets)]
    return sum(outs) / len(outputs) * 100


def computePartialEM(outputs, targets):
    outs = [metric_max_over_ground_truths(partial_exact_match, o, t) for o, t in zip(outputs, targets)]
    return sum(outs) / len(outputs) * 100


def computeSM(outputs, targets):
    outs = [metric_max_over_ground_truths(structure_match, o, t) for o, t in zip(outputs, targets)]
    return sum(outs) / len(outputs) * 100


def computeBLEU(outputs, targets):
    targets = [[t[i] for t in targets] for i in range(len(targets[0]))]
    return corpus_bleu(outputs, targets, lowercase=True).score


class Rouge(Rouge155):
    """Rouge calculator class with custom command-line options."""

    # See full list of options here:
    # https://github.com/andersjo/pyrouge/blob/master/tools/ROUGE-1.5.5/README.txt#L82
    DEFAULT_OPTIONS = [
        '-a',  # evaluate all systems
        '-n',
        4,  # max-ngram
        '-x',  # do not calculate ROUGE-L
        '-2',
        4,  # max-gap-length
        '-u',  # include unigram in skip-bigram
        '-c',
        95,  # confidence interval
        '-r',
        1000,  # number-of-samples (for resampling)
        '-f',
        'A',  # scoring formula
        '-p',
        0.5,  # 0 <= alpha <=1
        '-t',
        0,  # count by token instead of sentence
        '-d',  # print per evaluation scores
    ]

    def __init__(self, n_words=None, keep_files=False, options=None):

        if options is None:
            self.options = self.DEFAULT_OPTIONS.copy()
        else:
            self.options = options

        if n_words:
            options.extend(["-l", n_words])

        stem = "-m" in self.options

        super(Rouge, self).__init__(n_words=n_words, stem=stem, keep_files=keep_files)

    def _run_rouge(self):
        # Get full options
        options = ['-e', self._rouge_data] + list(map(str, self.options)) + [os.path.join(self._config_dir, "settings.xml")]

        # logging.info("Running ROUGE with options {}".format(" ".join(options)))
        # print([self._rouge_bin] + list(options))
        pipes = Popen([self._rouge_bin] + options, stdout=PIPE, stderr=PIPE)
        std_out, std_err = pipes.communicate()

        div_by_zero_error = std_err.decode("utf-8").startswith("Illegal division by zero")
        if pipes.returncode == 0 or div_by_zero_error:
            # Still returns the correct output even with div by zero
            return std_out
        else:
            raise ValueError(std_out.decode("utf-8") + "\n" + std_err.decode("utf-8"))


def computeROUGE(greedy, answer):
    rouges = compute_rouge_scores(greedy, answer)
    if len(rouges) > 0:
        avg_rouges = {}
        for key in rouges[0].keys():
            avg_rouges[key] = sum([r.get(key, 0.0) for r in rouges]) / len(rouges) * 100
    else:
        avg_rouges = None
    return avg_rouges


def split_sentences(txt, splitchar=".", include_splitchar=False):
    """Split sentences of a text based on a given EOS char."""
    out = [s.split() for s in txt.strip().split(splitchar) if len(s) > 0]
    return out


def compute_rouge_scores(summs, refs, splitchar='.', options=None, parallel=True):
    assert len(summs) == len(refs)
    options = [
        '-a',  # evaluate all systems
        '-c',
        95,  # confidence interval
        '-m',  # use Porter stemmer
        '-n',
        2,  # max-ngram
        '-w',
        1.3,  # weight (weighting factor for WLCS)
    ]
    rr = Rouge(options=options)
    rouge_args = []
    for summ, ref in zip(summs, refs):
        letter = "A"
        ref_dict = {}
        for r in ref:
            ref_dict[letter] = [x for x in split_sentences(r, splitchar) if len(x) > 0]
            letter = chr(ord(letter) + 1)
        s = [x for x in split_sentences(summ, splitchar) if len(x) > 0]
        rouge_args.append((s, ref_dict))
    if parallel:
        with closing(Pool(cpu_count() // 2)) as pool:
            rouge_scores = pool.starmap(rr.score_summary, rouge_args)
    else:
        rouge_scores = []
        for s, a in rouge_args:
            rouge_scores.append(rr.score_summary(s, ref_dict))
    return rouge_scores


def to_delta_state(line):
    delta_state = {'inform': {}, 'request': {}}
    try:
        if line == 'None' or line.strip() == '' or line.strip() == ';':
            return delta_state
        inform, request = [[y.strip() for y in x.strip().split(',')] for x in line.split(';')]
        inform_pairs = {}
        for i in inform:
            try:
                k, v = i.split(':')
                inform_pairs[k.strip()] = v.strip()
            except BaseException:
                pass
        delta_state = {'inform': inform_pairs, 'request': request}
    except BaseException:
        pass
    finally:
        return delta_state


def update_state(state, delta):
    for act, slot in delta.items():
        state[act] = slot
    return state


def dict_cmp(d1, d2):
    def cmp(a, b):
        for k1, v1 in a.items():
            if k1 not in b:
                return False
            else:
                if v1 != b[k1]:
                    return False
        return True

    return cmp(d1, d2) and cmp(d2, d1)


def computeDialogue(greedy, answer):
    examples = []
    for idx, (g, a) in enumerate(zip(greedy, answer)):
        examples.append((a[0][0], g, a[0][1], idx))
    examples.sort()
    turn_request_positives = 0
    turn_goal_positives = 0
    joint_goal_positives = 0
    ldt = None
    for ex in examples:
        if ldt is None or ldt.split('_')[:-1] != ex[0].split('_')[:-1]:
            state, answer_state = {}, {}
            ldt = ex[0]
        delta_state = to_delta_state(ex[1])
        answer_delta_state = to_delta_state(ex[2])
        state = update_state(state, delta_state['inform'])
        answer_state = update_state(answer_state, answer_delta_state['inform'])
        if dict_cmp(state, answer_state):
            joint_goal_positives += 1
        if delta_state['request'] == answer_delta_state['request']:
            turn_request_positives += 1
        if dict_cmp(delta_state['inform'], answer_delta_state['inform']):
            turn_goal_positives += 1

    joint_goal_em = joint_goal_positives / len(examples) * 100
    turn_request_em = turn_request_positives / len(examples) * 100
    turn_goal_em = turn_goal_positives / len(examples) * 100
    answer = [(x[-1], x[-2]) for x in examples]
    answer.sort()
    answer = [[x[1]] for x in answer]
    return joint_goal_em, turn_request_em, turn_goal_em, answer


def compute_metrics(greedy, answer, requested_metrics: Iterable):
    """
    Inputs:
        requested_metrics: contains a subset of the following metrics
            em (exact match)
            sm (structure match): valid if the output is ThingTalk code. Whether the gold answer and prediction are identical if we ignore parameter values of ThingTalk programs
            bleu
            rouge1, rouge2, rougeL, avg_rouge
            f1: token-level F1 score, tokenizes with whitespace
            nf1: normalize outputs then calculate token-level F1 score
            nem: normalize outputs then calculate exact match
            corpus_f1, precision, recall: corpus-level precision, recall and F1 score
            lfem
            joint_goal_em, turn_request_em, turn_goal_em, avg_dialogue
    """
    metric_keys = []
    metric_values = []
    if not isinstance(answer[0], list):
        answer = [[a] for a in answer]
    if 'lfem' in requested_metrics:
        lfem, answer = computeLFEM(greedy, answer)
        metric_keys += ['lfem']
        metric_values += [lfem]
    if 'joint_goal_em' in requested_metrics:
        joint_goal_em, request_em, turn_goal_em, answer = computeDialogue(greedy, answer)
        avg_dialogue = (joint_goal_em + request_em) / 2
        metric_keys += ['joint_goal_em', 'turn_request_em', 'turn_goal_em', 'avg_dialogue']
        metric_values += [joint_goal_em, request_em, turn_goal_em, avg_dialogue]
    em = computeEM(greedy, answer)
    metric_keys += ['em']
    metric_values += [em]
    if 'pem' in requested_metrics:
        pem = computePartialEM(greedy, answer)
        metric_keys.append('pem')
        metric_values.append(pem)
    if 'sm' in requested_metrics:
        sm = computeSM(greedy, answer)
        metric_keys.append('sm')
        metric_values.append(sm)
    if 'bleu' in requested_metrics:
        bleu = computeBLEU(greedy, answer)
        metric_keys.append('bleu')
        metric_values.append(bleu)
    if 'avg_rouge' in requested_metrics:
        rouge = computeROUGE(greedy, answer)
        metric_keys += ['rouge1', 'rouge2', 'rougeL', 'avg_rouge']
        avg_rouge = (rouge['rouge_1_f_score'] + rouge['rouge_2_f_score'] + rouge['rouge_l_f_score']) / 3
        metric_values += [rouge['rouge_1_f_score'], rouge['rouge_2_f_score'], rouge['rouge_l_f_score'], avg_rouge]
    if 'f1' in requested_metrics:
        f1 = computeF1(greedy, answer)
        metric_keys.append('f1')
        metric_values.append(f1)

    if 'ner_f1_IOB1' in requested_metrics:
        greedy_processed = [pred.split() for pred in greedy]
        answer_processed = [ans[0].split() for ans in answer]

        def convert_IOB2_to_IOB1(labels):
            cur_category = None
            for n, label in enumerate(labels):
                if label[0] == "B" and label[2:] != cur_category:
                    labels[n] = "I" + label[1:]
                cur_category = label[2:]

        convert_IOB2_to_IOB1(greedy_processed)
        convert_IOB2_to_IOB1(answer_processed)
        f1 = (
            seq_metrics.f1_score(y_pred=greedy_processed, y_true=answer_processed, mode='strict', scheme=seq_scheme.IOB1) * 100
        )

        metric_keys.append('ner_f1_IOB1')
        metric_values.append(f1)

    if 'ner_f1' in requested_metrics:
        greedy_processed = [pred.split() for pred in greedy]
        answer_processed = [ans[0].split() for ans in answer]

        f1 = seq_metrics.f1_score(y_pred=greedy_processed, y_true=answer_processed) * 100

        metric_keys.append('ner_f1')
        metric_values.append(f1)

    norm_greedy = [normalize_text(g) for g in greedy]
    norm_answer = [[normalize_text(a) for a in al] for al in answer]
    if 'nf1' in requested_metrics:
        nf1 = computeF1(norm_greedy, norm_answer)
        metric_keys.append('nf1')
        metric_values.append(nf1)
    if 'nem' in requested_metrics:
        nem = computeEM(norm_greedy, norm_answer)
        metric_keys.append('nem')
        metric_values.append(nem)

    if 'corpus_f1' in requested_metrics:
        corpus_f1, precision, recall = computeCF1(norm_greedy, norm_answer)
        metric_keys += ['corpus_f1', 'precision', 'recall']
        metric_values += [corpus_f1, precision, recall]

    metric_dict = dict(zip(metric_keys, metric_values))
    metric_dict = collections.OrderedDict((key, metric_dict[key]) for key in requested_metrics)
    return metric_dict, answer



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
from torch import nn
from torch.nn import functional as F
from transformers.modeling_outputs import Seq2SeqLMOutput

from .common import EPSILON, CombinedEmbedding, Feedforward, LSTMDecoderAttention, MultiLSTMCell, mask


class MQANDecoder(nn.Module):
    def __init__(self, numericalizer, args):
        super().__init__()
        self.numericalizer = numericalizer
        self.pad_idx = numericalizer.pad_id
        self.init_idx = numericalizer.init_id
        self.args = args

        self.decoder_embeddings = CombinedEmbedding(
            numericalizer, [], args.dimension, trained_dimension=args.trainable_decoder_embeddings, project=True
        )

        if args.rnn_layers > 0:
            self.rnn_decoder = LSTMDecoder(
                args.dimension, args.rnn_dimension, dropout=args.dropout_ratio, num_layers=args.rnn_layers
            )
            switch_input_len = 2 * args.rnn_dimension + args.dimension
        else:
            self.context_attn = LSTMDecoderAttention(args.dimension, dot=True)
            self.dropout = nn.Dropout(args.dropout_ratio)
            switch_input_len = 2 * args.dimension
        self.vocab_pointer_switch = nn.Sequential(Feedforward(switch_input_len, 1), nn.Sigmoid())

        self.generative_vocab_size = numericalizer.generative_vocab_size
        self.out = nn.Linear(args.rnn_dimension if args.rnn_layers > 0 else args.dimension, self.generative_vocab_size)

    def set_embeddings(self, embeddings):
        if self.decoder_embeddings is not None:
            self.decoder_embeddings.set_embeddings(embeddings)

    def forward(
        self,
        batch,
        final_context,
        context_rnn_state,
        encoder_loss,
        current_token_id=None,
        decoder_wrapper=None,
        expansion_factor=1,
        generation_dict=None,
    ):

        context, context_limited = batch.context.value, batch.context.limited
        answer, answer_limited = batch.answer.value, batch.answer.limited
        decoder_vocab = self.numericalizer.decoder_vocab
        self.map_to_full = decoder_vocab.decode
        context_padding = context.data == self.pad_idx
        if self.training:
            if self.args.rnn_layers > 0:
                self.rnn_decoder.applyMasks(context_padding)
            else:
                self.context_attn.applyMasks(context_padding)

            answer_padding = (answer.data == self.pad_idx)[:, :-1]
            answer_embedded = self.decoder_embeddings(answer[:, :-1], padding=answer_padding)

            if self.args.rnn_layers > 0:
                rnn_decoder_outputs = self.rnn_decoder(answer_embedded, final_context, hidden=context_rnn_state)
                decoder_output, vocab_pointer_switch_input, context_attention, rnn_state = rnn_decoder_outputs
            else:
                context_decoder_output, context_attention = self.context_attn(answer_embedded, final_context)
                vocab_pointer_switch_input = torch.cat((context_decoder_output, answer_embedded), dim=-1)
                decoder_output = self.dropout(context_decoder_output)

            vocab_pointer_switch = self.vocab_pointer_switch(vocab_pointer_switch_input)

            probs = self.probs(decoder_output, vocab_pointer_switch, context_attention, context_limited, decoder_vocab)

            probs, targets = mask(answer_limited[:, 1:].contiguous(), probs.contiguous(), pad_idx=decoder_vocab.pad_idx)
            loss = F.nll_loss(probs.log(), targets, ignore_index=decoder_vocab.pad_idx)
            if encoder_loss is not None:
                loss += self.args.encoder_loss_weight * encoder_loss

            return Seq2SeqLMOutput(loss=loss)
        else:
            if decoder_wrapper is None:
                decoder_wrapper = self.decoder_wrapper(
                    final_context,
                    context_padding,
                    context_limited,
                    decoder_vocab,
                    rnn_state=context_rnn_state,
                    expansion_factor=expansion_factor,
                    generation_dict=generation_dict,
                )
            else:
                current_token_id = current_token_id.clone().cpu().apply_(self.map_to_full).to(current_token_id.device)
            # (next_token_logits, past) where `past` includes all the states needed to continue generation
            logits = torch.log(decoder_wrapper.next_token_probs(current_token_id))
            return Seq2SeqLMOutput(logits=logits, past_key_values=decoder_wrapper)

    def probs(self, outputs, vocab_pointer_switches, context_attention, context_indices, decoder_vocab):
        size = list(outputs.size())

        size[-1] = self.generative_vocab_size
        scores = self.out(outputs.view(-1, outputs.size(-1))).view(size)
        p_vocab = F.softmax(scores, dim=scores.dim() - 1)
        scaled_p_vocab = vocab_pointer_switches.expand_as(p_vocab) * p_vocab

        effective_vocab_size = len(decoder_vocab)
        if self.generative_vocab_size < effective_vocab_size:
            size[-1] = effective_vocab_size - self.generative_vocab_size
            buff = scaled_p_vocab.new_full(size, EPSILON)
            scaled_p_vocab = torch.cat([scaled_p_vocab, buff], dim=buff.dim() - 1)

        # p_context_ptr
        scaled_p_vocab.scatter_add_(
            scaled_p_vocab.dim() - 1,
            context_indices.unsqueeze(1).expand_as(context_attention),
            ((1 - vocab_pointer_switches)).expand_as(context_attention) * context_attention,
        )

        return scaled_p_vocab

    def decoder_wrapper(
        self,
        context,
        context_padding,
        context_indices,
        decoder_vocab,
        rnn_state=None,
        expansion_factor=1,
        generation_dict=None,
    ):
        batch_size = context.size()[0]
        max_decoder_time = generation_dict['max_output_length']

        decoder_wrapper = MQANDecoderWrapper(
            context,
            context_padding,
            context_indices,
            decoder_vocab,
            rnn_state,
            batch_size,
            max_decoder_time,
            self,
            expansion_factor=expansion_factor,
        )

        return decoder_wrapper


class LSTMDecoder(nn.Module):
    def __init__(self, d_in, d_hid, dropout=0.0, num_layers=1):
        super().__init__()
        self.d_hid = d_hid
        self.d_in = d_in
        self.num_layers = num_layers
        self.dropout = nn.Dropout(dropout)

        self.input_feed = True
        if self.input_feed:
            d_in += 1 * d_hid

        self.rnn = MultiLSTMCell(self.num_layers, d_in, d_hid, dropout)
        self.context_attn = LSTMDecoderAttention(d_hid, dot=True)

    def applyMasks(self, context_mask):
        self.context_attn.applyMasks(context_mask)

    def forward(self, input: torch.Tensor, context, output=None, hidden=None):
        context_output = output if output is not None else self.make_init_output(context)

        context_outputs, vocab_pointer_switch_inputs, context_attentions = [], [], []
        for decoder_input in input.split(1, dim=1):
            context_output = self.dropout(context_output)
            if self.input_feed:
                rnn_input = torch.cat([decoder_input, context_output], 2)
            else:
                rnn_input = decoder_input

            rnn_input = rnn_input.squeeze(1)
            dec_state, hidden = self.rnn(rnn_input, hidden)
            dec_state = dec_state.unsqueeze(1)

            context_output, context_attention = self.context_attn(dec_state, context)
            vocab_pointer_switch_inputs.append(torch.cat([dec_state, context_output, decoder_input], -1))

            context_output = self.dropout(context_output)
            context_outputs.append(context_output)
            context_attentions.append(context_attention)

        return [torch.cat(x, dim=1) for x in (context_outputs, vocab_pointer_switch_inputs, context_attentions)] + [hidden]

    def make_init_output(self, context):
        batch_size = context.size(0)
        h_size = (batch_size, 1, self.d_hid)
        return context.new_zeros(h_size)


class MQANDecoderWrapper(object):
    """
    A wrapper for MQANDecoder that wraps around its recurrent neural network, so that we can decode it like a Transformer
    """

    def __init__(
        self,
        context,
        context_padding,
        context_indices,
        decoder_vocab,
        rnn_state,
        batch_size,
        max_decoder_time,
        mqan_decoder: MQANDecoder,
        expansion_factor: int,
    ):
        self.decoder_vocab = decoder_vocab
        context = self.expand_for_beam_search(context, batch_size, expansion_factor)
        context_padding = self.expand_for_beam_search(context_padding, batch_size, expansion_factor)
        context_indices = self.expand_for_beam_search(context_indices, batch_size, expansion_factor)
        if rnn_state is not None:
            rnn_state = self.expand_for_beam_search(rnn_state, batch_size, expansion_factor, dim=1)
        self.context = context
        self.context_padding = context_padding
        self.context_indices = context_indices
        self.rnn_state = rnn_state
        self.batch_size = batch_size
        self.max_decoder_time = max_decoder_time
        self.mqan_decoder = mqan_decoder

        if self.mqan_decoder.args.rnn_layers > 0:
            self.mqan_decoder.rnn_decoder.applyMasks(self.context_padding)
        else:
            self.mqan_decoder.context_attn.applyMasks(self.context_padding)

        self.time = 0
        self.decoder_output = None

    def reorder(self, new_order):
        # TODO only reordering rnn_state should be enough since reordering happens among beams of the same input
        self.context = self.reorder_for_beam_search(self.context, new_order)
        self.context_padding = self.reorder_for_beam_search(self.context_padding, new_order)
        self.context_indices = self.reorder_for_beam_search(self.context_indices, new_order)
        self.rnn_state = self.reorder_for_beam_search(self.rnn_state, new_order, dim=1)

        if self.mqan_decoder.args.rnn_layers > 0:
            self.mqan_decoder.rnn_decoder.applyMasks(self.context_padding)
        else:
            self.mqan_decoder.context_attn.applyMasks(self.context_padding)

    def next_token_probs(self, current_token_id):
        embedding = self.mqan_decoder.decoder_embeddings(current_token_id)

        if self.mqan_decoder.args.rnn_layers > 0:
            rnn_decoder_outputs = self.mqan_decoder.rnn_decoder(
                embedding, self.context, hidden=self.rnn_state, output=self.decoder_output
            )
            self.decoder_output, vocab_pointer_switch_input, context_attention, self.rnn_state = rnn_decoder_outputs
        else:
            context_decoder_output, context_attention = self.mqan_decoder.context_attn(embedding, self.context)
            vocab_pointer_switch_input = torch.cat((context_decoder_output, embedding), dim=-1)

            self.decoder_output = self.mqan_decoder.dropout(context_decoder_output)

        vocab_pointer_switch = self.mqan_decoder.vocab_pointer_switch(vocab_pointer_switch_input)

        probs = self.mqan_decoder.probs(
            self.decoder_output, vocab_pointer_switch, context_attention, self.context_indices, self.decoder_vocab
        )

        self.time += 1
        return probs

    def expand_for_beam_search(self, t, batch_size, num_beams, dim=0):
        if isinstance(t, tuple):
            elements = []
            for e in t:
                elements.append(self.expand_for_beam_search(e, batch_size, num_beams, dim))
            return tuple(elements)
        elif isinstance(t, list):
            elements = []
            for e in t:
                elements.append(self.expand_for_beam_search(e, batch_size, num_beams, dim))
            return elements

        original_size = list(t.shape)
        original_size[dim] *= num_beams
        t = t.unsqueeze(dim + 1)
        expanded_size = list(t.shape)
        expanded_size[dim + 1] = num_beams
        t = t.expand(*expanded_size)
        t = t.contiguous().view(*original_size)  # (batch_size * num_beams, -1)
        return t

    def reorder_for_beam_search(self, t, new_order, dim=0):
        if isinstance(t, tuple):
            elements = []
            for e in t:
                elements.append(self.reorder_for_beam_search(e, new_order, dim))
            return tuple(elements)
        elif isinstance(t, list):
            elements = []
            for e in t:
                elements.append(self.reorder_for_beam_search(e, new_order, dim))
            return elements

        p = [i for i in range(len(t.shape))]
        p[dim] = 0
        p[0] = dim
        t = t.permute(*p)
        t = t[new_order]
        t = t.permute(*p)

        return t
